from triage.component.architect.features import FeatureGroupCreator
from unittest import TestCase


def test_table_group():
    group = FeatureGroupCreator(
        definition={'tables': ['one', 'three']}
    )

    subsets = group.subsets({
        'one': ['col_a', 'col_b', 'col_c'],
        'two': ['col_a', 'col_b', 'col_c'],
        'three': ['col_a', 'col_b', 'col_c'],
    })
    assert subsets == [
        {'one': ['col_a', 'col_b', 'col_c']},
        {'three': ['col_a', 'col_b', 'col_c']},
    ]
    assert subsets[0].names == ['tables: one']
    assert subsets[1].names == ['tables: three']


def test_prefix_group():
    # ensure we test prefixes with underscores
    group = FeatureGroupCreator(
        definition={'prefix': ['major_viol', 'severe_viol']}
    )

    subsets = group.subsets({
        'one': ['minor_viol_a', 'minor_viol_b', 'minor_viol_c'],
        'two': ['severe_viol_a', 'severe_viol_b', 'severe_viol_c'],
        'three': ['major_viol_a', 'major_viol_b', 'major_viol_c'],
        'four': ['minor_viol_a', 'minor_viol_b', 'minor_viol_c'],
    })
    assert subsets == [
        {'three': ['major_viol_a', 'major_viol_b', 'major_viol_c']},
        {'two': ['severe_viol_a', 'severe_viol_b', 'severe_viol_c']},
    ]
    assert subsets[0].names == ['prefix: major_viol']
    assert subsets[1].names == ['prefix: severe_viol']


def test_multiple_criteria():
    group = FeatureGroupCreator(
        definition={
            'prefix': ['major', 'severe'],
            'tables': ['one', 'two'],
        }
    )

    subsets = group.subsets({
        'one': ['minor_a', 'minor_b', 'minor_c'],
        'two': ['severe_a', 'severe_b', 'severe_c'],
        'three': ['major_a', 'major_b', 'major_c'],
        'four': ['minor_a', 'minor_b', 'minor_c'],
    })
    assert subsets == [
        {'three': ['major_a', 'major_b', 'major_c']},
        {'two': ['severe_a', 'severe_b', 'severe_c']},
        {'one': ['minor_a', 'minor_b', 'minor_c']},
        {'two': ['severe_a', 'severe_b', 'severe_c']},
    ]

    assert subsets[0].names == ['prefix: major']
    assert subsets[1].names == ['prefix: severe']
    assert subsets[2].names == ['tables: one']
    assert subsets[3].names == ['tables: two']


class TestValidations(TestCase):
    def test_not_a_dict(self):
        creator = FeatureGroupCreator(definition='prefix')
        with self.assertRaises(ValueError):
            creator.validate()

    def test_badsubsetter(self):
        creator = FeatureGroupCreator(definition={
            'prefix': ['major', 'severe'],
            'other': ['other'],
        })
        with self.assertRaises(ValueError):
            creator.validate()

    def test_stringvalue(self):
        creator = FeatureGroupCreator(definition={
            'prefix': 'major',
        })
        with self.assertRaises(ValueError):
            creator.validate()

    def test_goodconfig(self):
        creator = FeatureGroupCreator(definition={
            'prefix': ['major', 'severe'],
            'tables': ['one', 'two'],
        })
        creator.validate()



"""Tests for Metta IO"""
import copy
import datetime
import os
import unittest
from dateutil.relativedelta import relativedelta
from shutil import rmtree
from tempfile import mkdtemp

import pandas as pd

from triage.component.metta import metta_io


DATA = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')

example_data_csv = os.path.join(DATA, 'titanic.csv')
example_data_h5 = os.path.join(DATA, 'titanic.h5')

dict_test_config = {
    'feature_start_time': datetime.date(2016, 1, 1),
    'end_time': datetime.date(2016, 12, 31),
    'matrix_id': 'testing_matrix',
    'label': 'testing_data',
    'label_name': 'Survived',
    'label_timespan': '1yr',
    'feature_names': [
        'break_last_3y',
        'soil',
        'pressure_zone',
    ],
}


class TestMettaIO(unittest.TestCase):
    """Tests Metta IO functionality"""

    def setUp(self):
        self.temp_dir = mkdtemp()

    def tearDown(self):
        rmtree(self.temp_dir)

    def temp_file(self, fname):
        return os.path.join(self.temp_dir, fname)

    def test_config(self):

        metta_io.check_config_types(dict_test_config)

    def test_uuid(self):
        fake_uuid = 'd1193048d65fca97ed10d494c49e2d1e'

        assert fake_uuid == metta_io.generate_uuid(dict_test_config)

    def test_store_matrix(self):
        df_data = pd.read_csv(example_data_csv)

        metta_io._store_matrix(
            dict_test_config,
            df_data,
            'test_titanic',
            self.temp_dir,
            format='csv'
        )
        metta_io._store_matrix(
            dict_test_config,
            df_data,
            'test_titanich5',
            self.temp_dir,
            format='hd5'
        )

        # check it wrote to files
        assert os.path.isfile(self.temp_file('test_titanic.csv'))
        assert os.path.isfile(self.temp_file('test_titanich5.h5'))
        assert os.path.isfile(self.temp_file('test_titanic.yaml'))
        assert os.path.isfile(self.temp_file('test_titanich5.yaml'))

    def test_archive_matrix(self):
        df_data = pd.read_csv(example_data_csv)

        train_uuid = metta_io.archive_matrix(
            dict_test_config,
            example_data_csv,
            directory=self.temp_dir,
            format='csv')

        train_uuid = metta_io.archive_matrix(
            dict_test_config, example_data_h5,
            directory=self.temp_dir, format='csv')

        train_uuid = metta_io.archive_matrix(
            dict_test_config, df_data,
            directory=self.temp_dir, format='csv')

        # check it wrote to files
        assert os.path.isfile(self.temp_file('{}.csv'.format(train_uuid)))
        assert os.path.isfile(self.temp_file('{}.yaml'.format(train_uuid)))

        def store_new_split(years):
            new_test_config = copy.deepcopy(dict_test_config)
            new_test_config['feature_start_time'] += relativedelta(years=years)
            new_test_config['end_time'] += relativedelta(years=years)
            return metta_io.archive_matrix(
                new_test_config,
                df_data,
                directory=self.temp_dir,
                format='csv',
                train_uuid=train_uuid
            )

        test_uuids = [store_new_split(years) for years in range(1, 5)]

        for test_uuid in test_uuids:
            assert os.path.isfile(self.temp_file('{}.csv'.format(test_uuid)))
            assert os.path.isfile(self.temp_file('{}.yaml'.format(test_uuid)))

    def test_archive_train_test(self):
        df_data = pd.read_csv(example_data_csv)

        metta_io.archive_train_test(dict_test_config,
                                    df_data,
                                    dict_test_config,
                                    df_data,
                                    directory=self.temp_dir,
                                    format='hd5',
                                    overwrite=False)

        # check that you don't write to a file again
        metta_io.archive_train_test(dict_test_config,
                                    df_data,
                                    dict_test_config,
                                    df_data,
                                    directory=self.temp_dir,
                                    format='hd5',
                                    overwrite=False)

        assert os.path.isfile(
            self.temp_file('d1193048d65fca97ed10d494c49e2d1e.h5')
        )

        assert os.path.isfile(
            self.temp_file('d1193048d65fca97ed10d494c49e2d1e.yaml')
        )

        prior_creation_time = os.path.getmtime(
            self.temp_file('d1193048d65fca97ed10d494c49e2d1e.h5'))

        metta_io.archive_train_test(dict_test_config,
                                    df_data,
                                    dict_test_config,
                                    df_data,
                                    directory=self.temp_dir,
                                    format='hd5',
                                    overwrite=True)

        later_creation_time = os.path.getmtime(
            self.temp_file('d1193048d65fca97ed10d494c49e2d1e.h5'))

        assert (later_creation_time - prior_creation_time) > 0

        assert len(os.listdir(self.temp_dir)) == 2

    def test_recover(self):
        df_data = pd.read_csv(example_data_csv)
        fake_uuid = 'd1193048d65fca97ed10d494c49e2d1e'
        metta_io.archive_train_test(dict_test_config, df_data,
                                    dict_test_config, df_data,
                                    directory=self.temp_dir)

        assert metta_io.recover_matrix('garbageuuid',
                                       directory=self.temp_dir) is None

        df_uuid = metta_io.recover_matrix(dict_test_config,
                                          directory=self.temp_dir)
        assert df_data.equals(df_uuid)

        df_config = metta_io.recover_matrix(fake_uuid,
                                            directory=self.temp_dir)
        assert df_data.equals(df_config)



import unittest
from triage.component.audition.selection_rules import BoundSelectionRule
from triage.component.audition.selection_rule_grid import make_selection_rule_grid

from triage.component.audition.rules_maker import (
    BaseRules,
    SimpleRuleMaker,
    RandomGroupRuleMaker,
    TwoMetricsRuleMaker,
    create_selection_grid
)

class TestSimpleRuleMaker(unittest.TestCase):
    def test_add_rule_best_current_value(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_current_value(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                'selection_rules': [{'name': 'best_current_value', 'n': 1}],
                'shared_parameters': [{'metric':'precision@', 'parameter':'100_abs'}]
            }
        ]

    def test_add_rule_best_average_value(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_average_value(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                'selection_rules': [{'name': 'best_average_value', 'n': 1}],
                'shared_parameters': [{'metric': 'precision@', 'parameter': '100_abs'}]
            }
        ]

    def test_add_rule_lowest_metric_variance(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_lowest_metric_variance(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                'selection_rules': [{'name': 'lowest_metric_variance', 'n': 1}],
                'shared_parameters': [{'metric': 'precision@', 'parameter': '100_abs'}]
            }
        ]

    def test_add_rule_most_frequent_best_dist(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_most_frequent_best_dist(metric="precision@", parameter="100_abs",
                                         dist_from_best_case=[0.01, 0.05, 0.1, 0.15])
        assert Rule.create() == [
            {
                'selection_rules': [
                    {
                        'dist_from_best_case': [0.01, 0.05, 0.1, 0.15],
                        'name': 'most_frequent_best_dist',
                        'n': 1
                    }
                ],
                'shared_parameters': [{'metric': 'precision@', 'parameter': '100_abs'}]
            }
        ]

    def test_add_rule_best_avg_recency_weight(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_avg_recency_weight(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                'selection_rules': [
                    {
                        'curr_weight': [1.5, 2.0, 5.0],
                        'decay_type': ['linear'],
                        'name': 'best_avg_recency_weight',
                        'n': 1
                    }
                ],
                'shared_parameters': [{'metric': 'precision@', 'parameter': '100_abs'}]
            }
        ]

    def test_add_rule_best_avg_var_penalized(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_avg_var_penalized(metric="precision@", parameter="100_abs", stdev_penalty=0.5)
        assert Rule.create() == [
            {
                'selection_rules': [
                    {
                        'name': 'best_avg_var_penalized',
                        'stdev_penalty': 0.5,
                        'n': 1
                    }
                ],
                'shared_parameters': [{'metric': 'precision@', 'parameter': '100_abs'}]
            }
        ]

class TestRandomGroupRuleMaker(unittest.TestCase):
    def test_random_model_groups(self):
        Rule = RandomGroupRuleMaker()
        assert Rule.create() == [
            {
                'selection_rules': [{'name': 'random_model_group', 'n': 1}],
                'shared_parameters': [{}]
            }
        ]

class TestTwoMetricsRuleMaker(unittest.TestCase):
    def test_add_two_metrics_rule_maker(self):
        Rule = TwoMetricsRuleMaker()
        Rule.add_rule_best_average_two_metrics(metric1='precision@', parameter1='100_abs',
            metric2='recall@', parameter2='300_abs', metric1_weight=[0.5])
        assert Rule.create() == [
            {
                'selection_rules': [
                    {
                        'metric1_weight': [0.5],
                        'name': 'best_average_two_metrics',
                        'metric2': ['recall@'],
                        'parameter2': ['300_abs'],
                        'n': 1
                    }
                ],
                'shared_parameters': [
                    {
                        'metric1': 'precision@',
                        'parameter1': '100_abs'
                    }
                ]
            }
        ]

class TestCreateSelectionRuleGrid(unittest.TestCase):
    def test_create_grid(self):
        """
        input_data = [{
            'shared_parameters': [
                {'metric': 'precision@', 'parameter': '100_abs'},
                {'metric': 'recall@', 'parameter': '100_abs'},
            ],
            'selection_rules': [
                {'name': 'most_frequent_best_dist', 'dist_from_best_case': [0.1, 0.2, 0.3]},
                {'name': 'best_current_value'}
            ]
        }, {
            'shared_parameters': [
                {'metric1': 'precision@', 'parameter1': '100_abs'},
            ],
            'selection_rules': [
                {
                    'name': 'best_average_two_metrics',
                    'metric2': ['recall@'],
                    'parameter2': ['100_abs'],
                    'metric1_weight': [0.4, 0.5, 0.6]
                },
            ]
        }]
        """
        Rule1 = SimpleRuleMaker()
        Rule1.add_rule_best_current_value(metric="precision@", parameter="100_abs")
        Rule1.add_rule_most_frequent_best_dist(metric="recall@",
                                               parameter="100_abs",
                                               dist_from_best_case=[0.1, 0.2, 0.3])

        Rule2 = TwoMetricsRuleMaker()
        Rule2.add_rule_best_average_two_metrics(metric1="precision@",
                                                parameter1="100_abs",
                                                metric2="recall@",
                                                parameter2="100_abs",
                                                metric1_weight=[0.4, 0.5, 0.6])


        expected_output = [
            BoundSelectionRule(
                descriptive_name='most_frequent_best_dist_precision@_100_abs_0.1',
                function_name='most_frequent_best_dist',
                args={'metric': 'precision@', 'parameter': '100_abs', 'dist_from_best_case': 0.1}
            ),
            BoundSelectionRule(
                descriptive_name='most_frequent_best_dist_precision@_100_abs_0.2',
                function_name='most_frequent_best_dist',
                args={'metric': 'precision@', 'parameter': '100_abs', 'dist_from_best_case': 0.2}
            ),
            BoundSelectionRule(
                descriptive_name='most_frequent_best_dist_precision@_100_abs_0.3',
                function_name='most_frequent_best_dist',
                args={'metric': 'precision@', 'parameter': '100_abs', 'dist_from_best_case': 0.3}
            ),
            BoundSelectionRule(
                descriptive_name='most_frequent_best_dist_recall@_100_abs_0.1',
                function_name='most_frequent_best_dist',
                args={'metric': 'recall@', 'parameter': '100_abs', 'dist_from_best_case': 0.1}
            ),
            BoundSelectionRule(
                descriptive_name='most_frequent_best_dist_recall@_100_abs_0.2',
                function_name='most_frequent_best_dist',
                args={'metric': 'recall@', 'parameter': '100_abs', 'dist_from_best_case': 0.2}
            ),
            BoundSelectionRule(
                descriptive_name='most_frequent_best_dist_recall@_100_abs_0.3',
                function_name='most_frequent_best_dist',
                args={'metric': 'recall@', 'parameter': '100_abs', 'dist_from_best_case': 0.3}
            ),
            BoundSelectionRule(
                descriptive_name='best_current_value_precision@_100_abs',
                function_name='best_current_value',
                args={'metric': 'precision@', 'parameter': '100_abs'}
            ),
            BoundSelectionRule(
                descriptive_name='best_current_value_recall@_100_abs',
                function_name='best_current_value',
                args={'metric': 'recall@', 'parameter': '100_abs'}
            ),
            BoundSelectionRule(
                descriptive_name='best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.4',
                function_name='best_average_two_metrics',
                args={
                    'metric1': 'precision@',
                    'parameter1': '100_abs',
                    'metric2': 'recall@',
                    'parameter2': '100_abs',
                    'metric1_weight': 0.4
                }
            ),
            BoundSelectionRule(
                descriptive_name='best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.5',
                function_name='best_average_two_metrics',
                args={
                    'metric1': 'precision@',
                    'parameter1': '100_abs',
                    'metric2': 'recall@',
                    'parameter2': '100_abs',
                    'metric1_weight': 0.5
                }
            ),
            BoundSelectionRule(
                descriptive_name='best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.6',
                function_name='best_average_two_metrics',
                args={
                    'metric1': 'precision@',
                    'parameter1': '100_abs',
                    'metric2': 'recall@',
                    'parameter2': '100_abs',
                    'metric1_weight': 0.6
                }
            )
        ]
        expected_output.sort(key=lambda x: x.descriptive_name)
        grid = sorted(make_selection_rule_grid(create_selection_grid(Rule1, Rule2)), key=lambda x: x.descriptive_name)
        assert len(grid) == len(expected_output)
        for expected_rule, actual_rule in zip(expected_output, grid):
            assert expected_rule.descriptive_name == actual_rule.descriptive_name



import datetime
import os
import random
import tempfile
from collections import OrderedDict
from contextlib import contextmanager
import pytest

import numpy
import pandas
import yaml
from sqlalchemy.orm import sessionmaker

from triage.component import metta
from triage.component.catwalk.storage import CSVMatrixStore, HDFMatrixStore, MatrixStore, ProjectStorage
from triage.component.results_schema import Model


@contextmanager
def fake_metta(matrix_dict, metadata):
    """Stores matrix and metadata in a metta-data-like form

    Args:
    matrix_dict (dict) of form { columns: values }.
        Expects an entity_id to be present which it will use as the index
    metadata (dict). Any metadata that should be set

    Yields:
        tuple of filenames for matrix and metadata
    """
    matrix = pandas.DataFrame.from_dict(matrix_dict).set_index('entity_id')
    with tempfile.NamedTemporaryFile() as matrix_file:
        with tempfile.NamedTemporaryFile('w') as metadata_file:
            hdf = pandas.HDFStore(matrix_file.name)
            hdf.put('title', matrix, data_columns=True)
            matrix_file.seek(0)

            yaml.dump(metadata, metadata_file)
            metadata_file.seek(0)
            yield (matrix_file.name, metadata_file.name)


def fake_labels(length):
    return numpy.array([random.choice([True, False]) for i in range(0, length)])


@pytest.fixture
def sample_metadata():
    return {
        'feature_start_time': datetime.date(2012, 12, 20),
        'end_time': datetime.date(2016, 12, 20),
        'label_name': 'label',
        'as_of_date_frequency': '1w',
        'max_training_history': '5y',
        'state': 'default',
        'cohort_name': 'default',
        'label_timespan': '1y',
        'metta-uuid': '1234',
        'feature_names': ['ft1', 'ft2'],
        'feature_groups': ['all: True'],
        'indices': ['entity_id'],
    }

@pytest.fixture
def sample_df():
    return pandas.DataFrame.from_dict({
        'entity_id': [1, 2],
        'feature_one': [3, 4],
        'feature_two': [5, 6],
        'label': ['good', 'bad']
    }).set_index('entity_id')

@pytest.fixture
def sample_matrix_store():
    with tempfile.TemporaryDirectory() as tempdir:
        project_storage = ProjectStorage(tempdir)
        store = project_storage.matrix_storage_engine().get_store('1234')
        store.matrix = sample_df()
        store.metadata = sample_metadata()
        return store



from triage.component.architect.feature_generators import FeatureGenerator
from triage.component.architect.feature_dictionary_creator import FeatureDictionaryCreator
from triage.component.architect.feature_group_creator import FeatureGroupCreator
from triage.component.architect.feature_group_mixer import FeatureGroupMixer

__all__ = (
    'FeatureGenerator',
    'FeatureDictionaryCreator',
    'FeatureGroupCreator',
    'FeatureGroupMixer',
)



import matplotlib.dates as md
import numpy as np
from triage.util.conf import convert_str_to_relativedelta
import matplotlib.pyplot as plt


FIG_SIZE = (32, 16)


def visualize_chops(chopper, show_as_of_times=True, show_boundaries=True):
    """Visualize time chops of a given Timechop object using matplotlib

    Args:
        chopper (triage.component.timechop.Timechop) A fully-configured Timechop object
        show_as_of_times (bool, default True) Whether or not to draw horizontal lines for as-of-times
        show_boundaries (bool, default True) Whether or not to show a rectangle around matrices
            and dashed lines around feature/label boundaries
    """
    chops = chopper.chop_time()

    chops.reverse()

    fig, ax = plt.subplots(len(chops), sharex=True, sharey=True, figsize=FIG_SIZE)

    for idx, chop in enumerate(chops):
        train_as_of_times = chop['train_matrix']['as_of_times']
        test_as_of_times = chop['test_matrices'][0]['as_of_times']

        test_label_timespan = chop['test_matrices'][0]['test_label_timespan']
        training_label_timespan = chop['train_matrix']['training_label_timespan']

        color_rgb = np.random.random(3)

        if(show_as_of_times):
            # Train matrix (as_of_times)
            ax[idx].hlines(
              [x for x in range(len(train_as_of_times))],
              [x.date() for x in train_as_of_times],
              [
                  x.date() + convert_str_to_relativedelta(training_label_timespan)
                  for x in train_as_of_times
              ],
              linewidth=3, color=color_rgb, label=f"train_{idx}"
            )

            # Test matrix
            ax[idx].hlines(
              [x for x in range(len(test_as_of_times))],
              [x.date() for x in test_as_of_times],
              [
                  x.date() + convert_str_to_relativedelta(test_label_timespan)
                  for x in test_as_of_times
              ],
              linewidth=3, color=color_rgb,
              label=f"test_{idx}"
            )

        if(show_boundaries):
            # Limits: train
            ax[idx].axvspan(
                chop['train_matrix']['first_as_of_time'],
                chop['train_matrix']['last_as_of_time'],
                color=color_rgb,
                alpha=0.3
            )

            ax[idx].axvline(chop['train_matrix']['matrix_info_end_time'], color='k', linestyle='--')

            # Limits: test
            ax[idx].axvspan(
                chop['test_matrices'][0]['first_as_of_time'],
                chop['test_matrices'][0]['last_as_of_time'],
                color=color_rgb,
                alpha=0.3
            )

            ax[idx].axvline(chop['feature_start_time'], color='k', linestyle='--', alpha=0.2)
            ax[idx].axvline(chop['feature_end_time'], color='k', linestyle='--',  alpha=0.2)
            ax[idx].axvline(chop['label_start_time'], color='k', linestyle='--', alpha=0.2)
            ax[idx].axvline(chop['label_end_time'], color='k', linestyle='--',  alpha=0.2)

            ax[idx].axvline(chop['test_matrices'][0]['matrix_info_end_time'], color='k', linestyle='--')

        ax[idx].yaxis.set_major_locator(plt.NullLocator())
        ax[idx].yaxis.set_label_position("right")
        ax[idx].set_ylabel(f"Block {idx}", rotation='horizontal', labelpad=30)

        ax[idx].xaxis.set_major_formatter(md.DateFormatter('%Y'))
        ax[idx].xaxis.set_major_locator(md.YearLocator())
        ax[idx].xaxis.set_minor_locator(md.MonthLocator())

    ax[0].set_title('Timechop: Temporal cross-validation blocks')
    fig.subplots_adjust(hspace=0)
    plt.setp([a.get_xticklabels() for a in fig.axes[:-1]], visible=False)
    plt.show()



from triage.component.timechop.utils import convert_to_list


def test_convert_to_list():
    tests = [
        {
            'val': '1 day',
            'expected_result': ['1 day']
        },
        {
            'val': ['1 day'],
            'expected_result': ['1 day']
        },
        {
            'val': 1,
            'expected_result': [1]
        }
    ]
    for test in tests:
        assert convert_to_list(test['val']) == test['expected_result']



"""Metta IO

Library for storing train-test sets.

"""
import copy
import yaml
import os
import datetime
import json
import hashlib
import shutil
import pandas as pd
import numpy as np

import logging

import s3fs

from urllib.parse import urlparse


def archive_train_test(train_config,
                       df_train,
                       test_config,
                       df_test,
                       directory='.',
                       format='hd5',
                       overwrite=False):
    """
    Main function for archiving train and test sets

    Parameters
    ----------
    (train_config, test_config): dict
        dicts to be yamled for train and test
    (df_train, df_test): df
        DataFrame of features and label (as last
        column) for training and testing set
    (title_train, title_test): str
        Unique name for train and test sets
    directory: str
        Relative path to where the data will be stored
    format: str
        format to save files in
        - hd5: HDF5
        - csv: Comma Separated Values
    overwrite: bool
        If true then identical matrices
        will be overridden.


    Returns
    -------
    (train_uuid, test_uuid): 'str'
       uuids for training and test set
    train_yaml: file
        Writes out to YAML file title.yaml
    data_file: file
        CSV of dataframe feature set title.csv

    """
    train_uuid = archive_matrix(
        train_config,
        df_train,
        overwrite=overwrite,
        directory=directory,
        format=format
    )

    test_uuid = archive_matrix(
        test_config,
        df_test,
        overwrite=overwrite,
        directory=directory,
        format=format,
        train_uuid=train_uuid
    )

    return train_uuid, test_uuid


def archive_matrix(
        matrix_config,
        df_matrix,
        overwrite=False,
        directory='.',
        format='hd5',
        train_uuid=None,
):
    """Store a design matrix.

    Parameters
    ----------
    matrix_config: dict
        dict to be yamled
    df_matrix: DataFrame or str
        DataFrame of features and label (as last column) or path to a CSV
    overwrite: bool
        If true will overwrite the same prexisting matrix.
    directory: str
        Relative path to where the data will be stored
    format: str
        format to save files in
        - hd5: HDF5
        - csv: Comma Separated Values
    train_uuid (optional): uuid of train set to associate with as a test set

    Returns
    -------
    uuid: str
        uuid for the stored set

    """
    if isinstance(df_matrix, pd.DataFrame):
        pass
    elif type(df_matrix) == str:
        abs_path_file = os.path.abspath(df_matrix)
        if not (os.path.isfile(abs_path_file)):
            raise IOError('Not a file: {}'.format(abs_path_file))

        if abs_path_file[-3:] == 'csv':
            pass
        elif abs_path_file[-2:] == 'h5':
            df_matrix = pd.read_hdf(abs_path_file)
        else:
            raise IOError('Not a csv or h5 file: {}'.format(abs_path_file))
    else:
        raise IOError(
            'Not a dataframe or path to a dataframe: {}'.type(df_matrix))

    check_config_types(matrix_config)

    matrix_uuid = generate_uuid(matrix_config)

    matrix_config = copy.deepcopy(matrix_config)
    matrix_config['metta-uuid'] = matrix_uuid

    fname = os.path.join(directory, matrix_uuid)

    # The output directory is local or in s3
    path_parsed = urlparse(directory)
    scheme = path_parsed.scheme  # If '' of 'file' is a regular file or 's3'

    if scheme in ('', 'file'):  # Local file
        abs_path_dir = os.path.abspath(directory)
        if not os.path.exists(abs_path_dir):
            os.makedirs(abs_path_dir)
        write_matrix = (overwrite) or not(os.path.exists(fname + format))
    elif scheme == 's3':
        abs_path_dir = directory
        write_matrix = (overwrite) or not(s3fs.S3FileSystem().exists(fname + format))
    else:
        raise ValueError(f"""URL scheme not supported:
              {scheme} (from {fname})
        """)

    if write_matrix:
        _store_matrix(matrix_config, df_matrix, matrix_uuid, abs_path_dir,
                      format=format)

    return matrix_uuid


def _store_matrix(metadata, df_data, title, directory, format='hd5'):
    """Store matrix and associated meta-data

    Parameters
    ----------
    metadata: dict
        dictionary of config/meta data
    data_df: df
        df of data with the last column being the labels
    title: str
        unique name of dataset
    directory: str
        Relative path to where the data will be stored
    format: str
        format to save files in
        - hd5: HDF5 (default) compressoin level 5, complib zlib
        - csv: Comma Separated Values

    Returns
    -------
    metadata: file
        Writes out to YAML file title.yaml
    data_file: file
        CSV of dataframe feature set title.csv

    Raises
    ------
    IOError:
        label name is not the last column name

    """
    # check last column is the label
    if isinstance(df_data, pd.DataFrame):
        last_col = df_data.columns.tolist()[-1]
    elif type(df_data) == str:
        abs_path_file = os.path.abspath(df_data)
        if not (os.path.isfile(abs_path_file)):
            raise IOError('Not a file: {}'.format(abs_path_file))
        if abs_path_file[-3:] == 'csv':
            headers = pd.read_csv(abs_path_file, nrows=1)
            last_col = headers.columns.tolist()[-1]

    if not (metadata['label_name'] == last_col):
        raise IOError('label_name is not last column')

    # The output directory is local or in s3
    path_parsed = urlparse(directory)
    scheme = path_parsed.scheme  # If '' of 'file' is a regular file or 's3'

    yaml_fname = os.path.join(directory, title + '.yaml')
    matrix_fname = os.path.join(directory, title + f".{format}")

    if scheme in ('', 'file'):  # Local file
        with open(yaml_fname, 'w') as stream:
            yaml.dump(metadata, stream)

        if format == 'hd5':  # H5 only supported locally
            matrix_fname = os.path.join(directory, title + ".h5")
            for col in df_data.columns:
                if (df_data[col].dtype == np.dtype('datetime64[ns]')):
                    df_data[col] = df_data[col].map(lambda x: x.timestamp())
                elif isinstance(df_data[col].dtype, object):
                    df_data[col] = df_data[col].astype(float)

            hdf = pd.HDFStore(matrix_fname,
                              mode='w',
                              complevel=5,
                              complib="zlib",
                              format='table')
            hdf.put(title, df_data, data_columns=True)
            hdf.close()
        elif format == 'csv':
            if isinstance(df_data, pd.DataFrame):
                if df_data.index.name:
                    df_data.to_csv(matrix_fname, index=False)
                else:
                    df_data.to_csv(matrix_fname)
            elif type(df_data) == str:
                if os.path.splitext(abs_path_file)[1] == 'csv':
                    shutil.copyfile(abs_path_file, matrix_fname)
        else:
            raise ValueError(f"""
                  File format not supported:
                  {format} (from {yaml_fname})
            """)
    elif scheme == 's3':
        s3 = s3fs.S3FileSystem()
        with s3.open(yaml_fname, "wb") as stream:
            yaml.dump(metadata, stream, encoding='utf-8')

        if isinstance(df_data, pd.DataFrame):
            logging.debug(f"Data frame received: {df_data.shape}")
            logging.debug(f"Data frame size in memory: " +
                          f"{df_data.memory_usage(index=True, deep=True).sum()} bytes")

            if df_data.index.name:
                bytes_to_write = df_data.to_csv(None, index=False).encode()
            else:
                bytes_to_write = df_data.to_csv(None).encode()
            with s3.open(matrix_fname, "wb") as stream:
                stream.write(bytes_to_write)

            logging.debug(f"Data frame stored in {matrix_fname}")
        elif type(df_data) == str:
            logging.debug(f"Local file received: {df_data}")
            s3.put(df_data, matrix_fname)
            logging.debug(f"Local file stored in {matrix_fname}")
        else:
            raise ValueError(f"""
                  Type not supported:
                  {type(df_data)}
            """)

    else:
        raise ValueError(f"""
              URL scheme not supported:
              {scheme} (from {yaml_fname})
        """)


def check_config_types(dict_config):
    """
    Enforce Datatypes in the dictionary


    Parameters
    ----------
    dict_config:
       Check the configuration types.
    The required types are:

    - feature_start_time: datetime.datetime
    - end_time: datetime.datetime
    - label_timespan: str
    - label_name: str
    - matrix_id: human readable name for the data

    Raises
    ------
    IOError:
        missing required keys

    """
    set_required_names = set(
        ['feature_start_time', 'end_time', 'label_timespan', 'label_name',
         'matrix_id'])

    if not (set_required_names.issubset(dict_config.keys())):
        raise IOError('missing required keys in dictionary',
                      set_required_names)

    # check that the start time and end times are correct
    assert isinstance(dict_config['feature_start_time'], datetime.date)
    assert isinstance(dict_config['end_time'], datetime.date)
    assert isinstance(dict_config['label_timespan'], str)
    assert isinstance(dict_config['matrix_id'], str)


def generate_uuid(metadata):
    """ Generate a unique identifier given a dictionary of matrix metadata.

    :param metadata: metadata for the matrix
    :type metadata: dict
    :returns: unique name for the file
    :rtype: str
    """

    def dt_handler(x):
        if isinstance(x, datetime.datetime) or isinstance(x, datetime.date):
            return x.isoformat()
        raise TypeError("Unknown type")

    return hashlib.md5(
        json.dumps(metadata, default=dt_handler, sort_keys=True)
            .encode('utf-8')
    ).hexdigest()


def recover_matrix(config, directory='.'):
    """Recover a matrix by either its config or uuid.

    Parameters
    ----------
    config: str or dict
        config metadata for the matrix or uuid
    directory: str
        path to search for the matrix

    Returns
    -------
    df_matrix: DataFrame
        DataFrame of specified matrix
    None:
        If no matrix matrix is found
    """

    if isinstance(config, dict):
        uuid = generate_uuid(config)
    else:
        uuid = config

    fname = directory + '/' + uuid

    if os.path.isfile(fname + '.h5'):
        df_matrix = pd.read_hdf(fname + '.h5')
        return df_matrix
    elif os.path.isfile(fname + '.csv'):
        df_matrix = pd.read_csv(fname + '.csv')
        return df_matrix
    else:
        return None



"""Testing Factories for creating database objects in unit tests

If init_engine is called first, objects are instantiated in the session at
module level. Example:

```
from .factories import EvaluationFactory, session, init_engine

engine = # your engine creation code here
init_engine(engine)
EvaluationFactory()
session.commit()

results = engine.execute('select * from results.evaluations')
```

"""
from datetime import datetime

import factory
import factory.fuzzy
from sqlalchemy.orm import sessionmaker, scoped_session

from triage.component import results_schema as schema


sessionmaker = sessionmaker()
session = scoped_session(sessionmaker)


class ExperimentFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.Experiment
        sqlalchemy_session = session
    experiment_hash = factory.fuzzy.FuzzyText()
    config = {}


class ModelGroupFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.ModelGroup
        sqlalchemy_session = session
    model_type = 'sklearn.ensemble.RandomForestClassifier'
    hyperparameters = {'hyperparam1': 'value1', 'hyperparam2': 'value2'}
    feature_list = ['feature1', 'feature2', 'feature3']
    model_config = {}

class MatrixFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.Matrix
        sqlalchemy_session = session
    matrix_uuid = factory.fuzzy.FuzzyText()

class ModelFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.Model
        sqlalchemy_session = session
    model_group_rel = factory.SubFactory(ModelGroupFactory)
    model_hash = factory.fuzzy.FuzzyText()
    run_time = factory.LazyFunction(lambda: datetime.now())
    batch_run_time = factory.LazyFunction(lambda: datetime.now())
    model_type = 'sklearn.ensemble.RandomForestClassifier'
    hyperparameters = {'hyperparam1': 'value1', 'hyperparam2': 'value2'}
    model_comment = ''
    batch_comment = ''
    config = {}
    matrix_rel = factory.SubFactory(MatrixFactory)
    experiment_rel = factory.SubFactory(ExperimentFactory)
    train_end_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    test = False
    train_matrix_uuid = factory.SelfAttribute('matrix_rel.matrix_uuid')
    training_label_timespan = '1y'

class FeatureImportanceFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.FeatureImportance
        sqlalchemy_session = session
    model_rel = factory.SubFactory(ModelFactory)
    feature = factory.fuzzy.FuzzyText()
    feature_importance = factory.fuzzy.FuzzyDecimal(0, 1)
    rank_abs = 1
    rank_pct = 1.0


class PredictionFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.TestPrediction
        sqlalchemy_session = session

    model_rel = factory.SubFactory(ModelFactory)
    entity_id = factory.fuzzy.FuzzyInteger(0)
    as_of_date = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    score = factory.fuzzy.FuzzyDecimal(0, 1)
    label_value = factory.fuzzy.FuzzyInteger(0, 1)
    rank_abs = 1
    rank_pct = 1.0
    matrix_uuid = factory.SelfAttribute('model_rel.train_matrix_uuid')
    test_label_timespan = '3m'


class ListPredictionFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.ListPrediction
        sqlalchemy_session = session

    model_rel = factory.SubFactory(ModelFactory)
    entity_id = factory.fuzzy.FuzzyInteger(0)
    as_of_date = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    score = factory.fuzzy.FuzzyDecimal(0, 1)
    rank_abs = 1
    rank_pct = 1.0
    matrix_uuid = "efgh"
    test_label_timespan = '3m'


class IndividualImportanceFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.IndividualImportance
        sqlalchemy_session = session
    model_rel = factory.SubFactory(ModelFactory)
    entity_id = factory.fuzzy.FuzzyInteger(0)
    as_of_date = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    feature = factory.fuzzy.FuzzyText()
    feature_value = factory.fuzzy.FuzzyDecimal(0, 100)
    method = factory.fuzzy.FuzzyText()
    importance_score = factory.fuzzy.FuzzyDecimal(0, 1)


class EvaluationFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.TestEvaluation
        sqlalchemy_session = session
    model_rel = factory.SubFactory(ModelFactory)
    evaluation_start_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    evaluation_end_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    as_of_date_frequency = '3d'
    metric = 'precision@'
    parameter = '100_abs'
    value = factory.fuzzy.FuzzyDecimal(0, 1)
    num_labeled_examples = 10
    num_labeled_above_threshold = 8
    num_positive_labels = 5
    sort_seed = 8


def init_engine(new_engine):
    global sessionmaker, engine, session
    engine = new_engine
    session.remove()
    sessionmaker.configure(bind=engine)



# -*- coding: utf-8 -*-
"""Integration tests for `collate` module."""
import testing.postgresql
from sqlalchemy import create_engine
from sqlalchemy.sql import expression as ex

from triage.component.collate import Aggregation, Aggregate
from triage.component.collate.spacetime import SpacetimeAggregation

from . import initialize_db


IMPUTE_RULES={
    'coltype': 'aggregate',
    'count': {'type': 'mean'},
    'mode': {'type': 'mean'}
}

Postgresql = testing.postgresql.PostgresqlFactory(cache_initialized_db=True,
                                                  on_initialized=initialize_db.handler)


def teardown_module():
    Postgresql.clear_cache()


def test_engine():
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ((result,),) = engine.execute("SELECT COUNT(*) FROM food_inspections")
    assert result == 966


def test_st_explicit_execute():
    agg = Aggregate({'F': "results='Fail'"}, ["count"], IMPUTE_RULES)
    mode = Aggregate("", "mode", IMPUTE_RULES, order="zip")
    st = SpacetimeAggregation(
        [agg, agg+agg, mode],
        from_obj = ex.table('food_inspections'),
        groups = {'license':ex.column('license_no'),
                  'zip':ex.column('zip')},
        intervals = {'license' : ["1 year", "2 years", "all"],
                     'zip' : ["1 year"]},
        dates = ['2016-08-30', '2015-11-06'],
        state_table = 'inspection_states',
        state_group = 'license_no',
        date_column = 'inspection_date',
        prefix='food_inspections'
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_st_lazy_execute():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = SpacetimeAggregation(
        [agg],
        from_obj = 'food_inspections',
        groups = ['license_no', 'zip'],
        intervals = {'license_no':["1 year", "2 years", "all"],
                           'zip' : ["1 year"]},
        dates = ['2016-08-30', '2015-11-06'],
        state_table = 'inspection_states',
        state_group = 'license_no',
        date_column = '"inspection_date"'
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_st_execute_broadcast_intervals():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = SpacetimeAggregation(
        [agg],
        from_obj = 'food_inspections',
        groups = ['license_no', 'zip'],
        intervals = ["1 year", "2 years", "all"],
        dates = ['2016-08-30', '2015-11-06'],
        state_table = 'inspection_states',
        state_group = 'license_no',
        date_column = '"inspection_date"'
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_execute():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = Aggregation(
        [agg],
        from_obj = 'food_inspections',
        groups = ['license_no', 'zip'],
        state_table = 'all_licenses',
        state_group = 'license_no',
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_execute_schema_output_date_column():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = SpacetimeAggregation(
        [agg],
        from_obj = 'food_inspections',
        groups = ['license_no', 'zip'],
        intervals = {'license_no':["1 year", "2 years", "all"],
                           'zip' : ["1 year"]},
        dates = ['2016-08-30', '2015-11-06'],
        state_table = 'inspection_states_diff_colname',
        state_group = 'license_no',
        schema = "agg",
        date_column = '"inspection_date"',
        output_date_column = "aggregation_date"
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())



from triage.component.catwalk.model_trainers import ModelTrainer
from triage.component.catwalk.predictors import Predictor
from triage.component.catwalk.evaluation import ModelEvaluator
from triage.component.catwalk.utils import save_experiment_and_get_hash
from triage.component.catwalk.db import ensure_db
from triage.component.catwalk.storage import ProjectStorage, ModelStorageEngine, MatrixStore
from tests.results_tests.factories import init_engine, session, MatrixFactory
from tests.utils import rig_engines, get_matrix_store, matrix_creator, matrix_metadata_creator

import boto3
import testing.postgresql

from moto import mock_s3
from sqlalchemy import create_engine
import datetime
import pandas
from .utils import sample_metadata


def test_integration():
    with rig_engines() as (db_engine, project_storage):
        train_store = get_matrix_store(
            project_storage,
            matrix_creator(),
            matrix_metadata_creator(matrix_type='train')
        )
        as_of_dates = [
            datetime.date(2016, 12, 21),
            datetime.date(2017, 1, 21)
        ]

        test_stores = []
        for as_of_date in as_of_dates:
            matrix_store = get_matrix_store(
                project_storage,
                pandas.DataFrame.from_dict(
                    {'entity_id': [3], 'feature_one': [8], 'feature_two': [5], 'label': [0]}
                ).set_index('entity_id'),
                matrix_metadata_creator(end_time=as_of_date, indices=['entity_id'])
            )
            test_stores.append(matrix_store)

        model_storage_engine = ModelStorageEngine(project_storage)

        experiment_hash = save_experiment_and_get_hash({}, db_engine)
        # instantiate pipeline objects
        trainer = ModelTrainer(
            experiment_hash=experiment_hash,
            model_storage_engine=model_storage_engine,
            db_engine=db_engine,
        )
        predictor = Predictor(
            model_storage_engine,
            db_engine
        )
        model_evaluator = ModelEvaluator(
            [{'metrics': ['precision@'], 'thresholds': {'top_n': [5]}}],
            [{}],
            db_engine
        )

        # run the pipeline
        grid_config = {
            'sklearn.linear_model.LogisticRegression': {
                'C': [0.00001, 0.0001],
                'penalty': ['l1', 'l2'],
                'random_state': [2193]
            }
        }
        model_ids = trainer.train_models(
            grid_config=grid_config,
            misc_db_parameters=dict(),
            matrix_store=train_store
        )

        for model_id in model_ids:
            for as_of_date, test_store in zip(as_of_dates, test_stores):
                predictions_proba = predictor.predict(
                    model_id,
                    test_store,
                    misc_db_parameters=dict(),
                    train_matrix_columns=['feature_one', 'feature_two']
                )

                model_evaluator.evaluate(
                    predictions_proba,
                    test_store,
                    model_id,
                )

        # assert
        # 1. that the predictions table entries are present and
        # can be linked to the original models
        records = [
            row for row in
            db_engine.execute('''select entity_id, model_id, as_of_date
            from test_results.predictions
            join model_metadata.models using (model_id)
            order by 3, 2''')
        ]
        assert records == [
            (3, 1, datetime.datetime(2016, 12, 21)),
            (3, 2, datetime.datetime(2016, 12, 21)),
            (3, 3, datetime.datetime(2016, 12, 21)),
            (3, 4, datetime.datetime(2016, 12, 21)),
            (3, 1, datetime.datetime(2017, 1, 21)),
            (3, 2, datetime.datetime(2017, 1, 21)),
            (3, 3, datetime.datetime(2017, 1, 21)),
            (3, 4, datetime.datetime(2017, 1, 21)),
        ]

        # that evaluations are there
        records = [
            row for row in
            db_engine.execute('''
                select model_id, evaluation_start_time, metric, parameter
                from test_results.evaluations order by 2, 1''')
        ]
        assert records == [
            (1, datetime.datetime(2016, 12, 21), 'precision@', '5_abs'),
            (2, datetime.datetime(2016, 12, 21), 'precision@', '5_abs'),
            (3, datetime.datetime(2016, 12, 21), 'precision@', '5_abs'),
            (4, datetime.datetime(2016, 12, 21), 'precision@', '5_abs'),
            (1, datetime.datetime(2017, 1, 21), 'precision@', '5_abs'),
            (2, datetime.datetime(2017, 1, 21), 'precision@', '5_abs'),
            (3, datetime.datetime(2017, 1, 21), 'precision@', '5_abs'),
            (4, datetime.datetime(2017, 1, 21), 'precision@', '5_abs'),
        ]



from triage.component.audition.selection_rules import BoundSelectionRule
from triage.component.audition.selection_rule_grid import make_selection_rule_grid


def test_selection_rule_grid():
    input_data = [{
        'shared_parameters': [
            {'metric': 'precision@', 'parameter': '100_abs'},
            {'metric': 'recall@', 'parameter': '100_abs'},
        ],
        'selection_rules': [
            {'name': 'most_frequent_best_dist', 'dist_from_best_case': [0.1, 0.2, 0.3]},
            {'name': 'best_current_value'}
        ]
    }, {
        'shared_parameters': [
            {'metric1': 'precision@', 'parameter1': '100_abs'},
        ],
        'selection_rules': [
            {
                'name': 'best_average_two_metrics',
                'metric2': ['recall@'],
                'parameter2': ['100_abs'],
                'metric1_weight': [0.4, 0.5, 0.6]
            },
        ]
    }]

    expected_output = [
        BoundSelectionRule(
            descriptive_name='most_frequent_best_dist_precision@_100_abs_0.1',
            function_name='most_frequent_best_dist',
            args={'metric': 'precision@', 'parameter': '100_abs', 'dist_from_best_case': 0.1}
        ),
        BoundSelectionRule(
            descriptive_name='most_frequent_best_dist_precision@_100_abs_0.2',
            function_name='most_frequent_best_dist',
            args={'metric': 'precision@', 'parameter': '100_abs', 'dist_from_best_case': 0.2}
        ),
        BoundSelectionRule(
            descriptive_name='most_frequent_best_dist_precision@_100_abs_0.3',
            function_name='most_frequent_best_dist',
            args={'metric': 'precision@', 'parameter': '100_abs', 'dist_from_best_case': 0.3}
        ),
        BoundSelectionRule(
            descriptive_name='most_frequent_best_dist_recall@_100_abs_0.1',
            function_name='most_frequent_best_dist',
            args={'metric': 'recall@', 'parameter': '100_abs', 'dist_from_best_case': 0.1}
        ),
        BoundSelectionRule(
            descriptive_name='most_frequent_best_dist_recall@_100_abs_0.2',
            function_name='most_frequent_best_dist',
            args={'metric': 'recall@', 'parameter': '100_abs', 'dist_from_best_case': 0.2}
        ),
        BoundSelectionRule(
            descriptive_name='most_frequent_best_dist_recall@_100_abs_0.3',
            function_name='most_frequent_best_dist',
            args={'metric': 'recall@', 'parameter': '100_abs', 'dist_from_best_case': 0.3}
        ),
        BoundSelectionRule(
            descriptive_name='best_current_value_precision@_100_abs',
            function_name='best_current_value',
            args={'metric': 'precision@', 'parameter': '100_abs'}
        ),
        BoundSelectionRule(
            descriptive_name='best_current_value_recall@_100_abs',
            function_name='best_current_value',
            args={'metric': 'recall@', 'parameter': '100_abs'}
        ),
        BoundSelectionRule(
            descriptive_name='best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.4',
            function_name='best_average_two_metrics',
            args={
                'metric1': 'precision@',
                'parameter1': '100_abs',
                'metric2': 'recall@',
                'parameter2': '100_abs',
                'metric1_weight': 0.4
            }
        ),
        BoundSelectionRule(
            descriptive_name='best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.5',
            function_name='best_average_two_metrics',
            args={
                'metric1': 'precision@',
                'parameter1': '100_abs',
                'metric2': 'recall@',
                'parameter2': '100_abs',
                'metric1_weight': 0.5
            }
        ),
        BoundSelectionRule(
            descriptive_name='best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.6',
            function_name='best_average_two_metrics',
            args={
                'metric1': 'precision@',
                'parameter1': '100_abs',
                'metric2': 'recall@',
                'parameter2': '100_abs',
                'metric1_weight': 0.6
            }
        )
    ]

    # sort both lists so we can compare them without resorting to a hash
    expected_output.sort(key=lambda x: x.descriptive_name)
    grid = sorted(make_selection_rule_grid(input_data), key=lambda x: x.descriptive_name)
    assert len(grid) == len(expected_output)
    for expected_rule, actual_rule in zip(expected_output, grid):
        assert expected_rule.descriptive_name == actual_rule.descriptive_name



import pandas
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.catwalk.db import ensure_db
from triage.component.catwalk.storage import MatrixStore
from triage.component.catwalk.individual_importance.uniform import uniform_distribution
from triage.component.results_schema import Matrix
from tests.utils import rig_engines, get_matrix_store, matrix_metadata_creator

from tests.results_tests.factories import (
    ModelFactory,
    FeatureImportanceFactory,
    init_engine,
    session,
    MatrixFactory
)


def test_uniform_distribution_entity_id_index():
    with rig_engines() as (db_engine, project_storage):
        model = ModelFactory()
        feature_importances = [
            FeatureImportanceFactory(model_rel=model, feature='feature_{}'.format(i))
            for i in range(0, 10)
        ]
        data_dict = {'entity_id': [1, 2]}
        for imp in feature_importances:
            data_dict[imp.feature] = [0.5, 0.5]
        metadata = matrix_metadata_creator(indices='entity_id')
        test_store = get_matrix_store(
            project_storage,
            pandas.DataFrame.from_dict(data_dict).set_index(metadata['indices']),
            metadata
        )
        results = uniform_distribution(
            db_engine,
            model_id=model.model_id,
            as_of_date='2016-01-01',
            test_matrix_store=test_store,
            n_ranks=5
        )

        assert len(results) == 10  # 5 features x 2 entities
        for result in results:
            assert 'entity_id' in result
            assert 'feature_name' in result
            assert 'score' in result
            assert 'feature_value' in result
            assert result['feature_value'] == 0.5
            assert result['score'] >= 0
            assert result['score'] <= 1
            assert isinstance(result['feature_name'], str)
            assert result['entity_id'] in [1, 2]


def test_uniform_distribution_entity_date_index():
    with rig_engines() as (db_engine, project_storage):
        model = ModelFactory()
        feature_importances = [
            FeatureImportanceFactory(model_rel=model, feature='feature_{}'.format(i))
            for i in range(0, 10)
        ]
        data_dict = {'entity_id': [1, 1], 'as_of_date': ['2016-01-01', '2017-01-01']}
        for imp in feature_importances:
            data_dict[imp.feature] = [0.5, 0.5]
        metadata = matrix_metadata_creator()
        test_store = get_matrix_store(
            project_storage,
            pandas.DataFrame.from_dict(data_dict).set_index(metadata['indices']),
            metadata
        )
        results = uniform_distribution(
            db_engine,
            model_id=model.model_id,
            as_of_date='2016-01-01',
            test_matrix_store=test_store,
            n_ranks=5
        )

        assert len(results) == 5  # 5 features x 1 entity for this as_of_date
        for result in results:
            assert 'entity_id' in result
            assert 'feature_name' in result
            assert 'score' in result
            assert 'feature_value' in result
            assert result['feature_value'] == 0.5
            assert result['score'] >= 0
            assert result['score'] <= 1
            assert isinstance(result['feature_name'], str)
            assert result['entity_id'] in [1, 2]



from unittest.mock import patch

import numpy
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.model_group_performance import ModelGroupPerformancePlotter

from .utils import create_sample_distance_table


def test_ModelGroupPerformancePlotter_generate_plot_data():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)
        plotter = ModelGroupPerformancePlotter(distance_table)
        df = plotter.generate_plot_data(
            metric='precision@',
            parameter='100_abs',
            model_group_ids=[1, 2],
            train_end_times=['2014-01-01', '2015-01-01'],
        )
        assert sorted(df['model_type'].unique()) == ['best case', 'mySpikeClassifier', 'myStableClassifier']
        for value in df[df['model_group_id'] == 1]['raw_value'].values:
            assert numpy.isclose(value, 0.5)


def test_ModelGroupPerformancePlotter_plot_all():
    with patch('triage.component.audition.model_group_performance.plot_cats') as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = ModelGroupPerformancePlotter(distance_table)
            plotter.plot_all(
                [{'metric': 'precision@', 'parameter': '100_abs'}],
                model_group_ids=[1, 2],
                train_end_times=['2014-01-01', '2015-01-01'],
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert 'raw_value' in kwargs['frame']
        assert 'train_end_time' in kwargs['frame']
        assert kwargs['x_col'] == 'train_end_time'
        assert kwargs['y_col'] == 'raw_value'



import testing.postgresql
import datetime
import pytest
from copy import copy

from sqlalchemy import create_engine
from triage.component.catwalk.db import ensure_db

from triage.component.catwalk.model_grouping import ModelGrouper
from .utils import sample_metadata


def test_model_grouping_default_config(sample_metadata):
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        model_grouper = ModelGrouper()
        # get the basic first model group with our default matrix
        assert model_grouper.get_model_group_id(
            'module.Classifier',
            {'param1': 'val1'},
            sample_metadata,
            engine
        ) == 1

        # the end time is not by default a model group key so changing it
        # should still get us the same group
        metadata_new_end_time = copy(sample_metadata)
        metadata_new_end_time['end_time'] = datetime.date(2017, 3, 20)
        assert model_grouper.get_model_group_id(
            'module.Classifier',
            {'param1': 'val1'},
            metadata_new_end_time,
            engine
        ) == 1

        # max_training_history is a default key,
        # so it should trigger a new group
        metadata_train_history = copy(sample_metadata)
        metadata_train_history['max_training_history'] = '3y'
        assert model_grouper.get_model_group_id(
            'module.Classifier',
            {'param1': 'val1'},
            metadata_train_history,
            engine
        ) == 2

        # classifier is of course a default key as well
        assert model_grouper.get_model_group_id(
            'module.OtherClassifier',
            {'param1': 'val1'},
            sample_metadata,
            engine
        ) == 3


def test_model_grouping_custom_config(sample_metadata):
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        model_grouper = ModelGrouper(model_group_keys=['feature_names', 'as_of_date_frequency'])
        # get the basic first model group with our default matrix
        assert model_grouper.get_model_group_id(
            'module.Classifier',
            {'param1': 'val1'},
            sample_metadata,
            engine
        ) == 1

        # classifier is now not a key, so changing it should not get a new id
        assert model_grouper.get_model_group_id(
            'module.OtherClassifier',
            {'param1': 'val1'},
            sample_metadata,
            engine
        ) == 1

        # as_of_date_frequency is a key,
        # so it should trigger a new group
        metadata_frequency = copy(sample_metadata)
        metadata_frequency['as_of_date_frequency'] = '2w'
        assert model_grouper.get_model_group_id(
            'module.Classifier',
            {'param1': 'val1'},
            metadata_frequency,
            engine
        ) == 2

        # testing feature names may seem redundant but it is on a separate
        # code path so make sure its logic works
        metadata_features = copy(sample_metadata)
        metadata_features['feature_names'] = ['ft1', 'ft3']
        assert model_grouper.get_model_group_id(
            'module.Classifier',
            {'param1': 'val1'},
            metadata_features,
            engine
        ) == 3



from triage.component.architect.features import FeatureDictionaryCreator
import testing.postgresql
from sqlalchemy import create_engine


def test_feature_dictionary_creator():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        engine.execute('create schema features')
        engine.execute('''
            create table features.prefix1_entity_id (
                entity_id int,
                as_of_date date,
                feature_one float,
                feature_two float
            )
        ''')
        engine.execute('''
            create table features.prefix1_zipcode (
                zipcode text,
                as_of_date date,
                feature_three float,
                feature_four float
            )
        ''')
        engine.execute('''
            create table features.prefix1_aggregation (
                entity_id int,
                as_of_date date,
                zipcode text,
                feature_one float,
                feature_two float,
                feature_three float,
                feature_four float
            )
        ''')
        engine.execute('''
            create table features.prefix1_aggregation_imputed (
                entity_id int,
                as_of_date date,
                zipcode text,
                feature_one float,
                feature_two float,
                feature_three float,
                feature_three_imp int,
                feature_four float
            )
        ''')
        engine.execute('''
            create table features.random_other_table (
                another_column float
            )
        ''')

        creator = FeatureDictionaryCreator(
            features_schema_name='features',
            db_engine=engine
        )
        feature_dictionary = creator.feature_dictionary(
            feature_table_names=[
                'prefix1_entity_id', 'prefix1_zip_code', 
                'prefix1_aggregation', 'prefix1_aggregation_imputed'
                ],
            index_column_lookup={
                'prefix1_aggregation_imputed': ['entity_id', 'zipcode', 'as_of_date']
            }
        )
        assert feature_dictionary == {
            'prefix1_aggregation_imputed': [
                'feature_one', 'feature_two', 'feature_three', 'feature_three_imp', 'feature_four'
                ],
        }



# -*- coding: utf-8 -*-
from .collate import (
    available_imputations,
    Aggregation,
    Aggregate,
    Compare,
    Categorical,
)
from .spacetime import SpacetimeAggregation

__all__ = [
    'available_imputations',
    'Aggregation',
    'Aggregate',
    'Compare',
    'Categorical',
    'SpacetimeAggregation',
]
__author__ = """DSaPP Researchers"""
__email__ = 'datascifellows@gmail.com'



import inspect
import logging

from numpy import exp, log, average

from .metric_directionality import greater_is_better, best_in_series, idxbest


def random_model_group(df, train_end_time, n=1):
    """Pick a random model group (as a baseline)

    Arguments:
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int) -- numbers of model group id
    Returns: (int) the model group id to select, with highest current raw metric value
    """
    return df['model_group_id'].drop_duplicates().sample(frac=1).tolist()[:n]


def _mg_best_avg_by(df, value_col, metric, n=1):
    """Best model group in dataframe by average of some column

    Args:
        df (pandas.DataFrame)
        value_col (str) The column which contains the value to be averaged
        metric (str) the name of the column
        n (int) numbers of model group id
    """
    if n == 1:
        return [getattr(
            df.groupby(['model_group_id'])[value_col].mean().sample(frac=1),
            idxbest(metric)
        )()]
    else:
        if greater_is_better(metric):
            return df.groupby(['model_group_id'])[value_col].mean().nlargest(n).index.tolist()
        else:
            return df.groupby(['model_group_id'])[value_col].mean().nsmallest(n).index.tolist()


def best_current_value(df, train_end_time, metric, parameter, n=1):
    """Pick the model group with the best current metric value

    Arguments:
        metric (string) -- model evaluation metric, such as 'precision@'
        parameter (string) -- model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                dist_from_best_case
        n (int) numbers of model group id
    Returns: (int) the model group id to select, with highest current raw metric value
    """
    curr_df = df.loc[
                (df['train_end_time'] == train_end_time) &
                (df['metric'] == metric) &
                (df['parameter'] == parameter)
              ]
    # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
    best_raw_value = getattr(curr_df['raw_value'], best_in_series(metric))()
    if n <= 1:
        return curr_df\
            .loc[curr_df['raw_value'] == best_raw_value, 'model_group_id']\
            .sample(frac=1)\
            .tolist()
    else:
        if greater_is_better(metric):
            result = curr_df.nlargest(n, 'raw_value')['model_group_id'].tolist()
            return result
        else:
            result = curr_df.nsmallest(n, 'raw_value')['model_group_id'].tolist()
            return result


def best_average_value(df, train_end_time, metric, parameter, n=1):
    """Pick the model with the highest average metric value so far

    Arguments:
        metric (string) -- model evaluation metric, such as 'precision@'
        parameter (string) -- model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                dist_from_best_case
        n (int) -- numbers of model group id
    Returns: (int) the model group id to select, with highest mean raw metric value
    """
    met_df = df.loc[
                (df['metric'] == metric) &
                (df['parameter'] == parameter)
            ]
    return _mg_best_avg_by(met_df, 'raw_value', metric, n)


def lowest_metric_variance(df, train_end_time, metric, parameter, n=1):
    """Pick the model with the lowest metric variance so far

    Arguments:
        metric (string) -- model evaluation metric, such as 'precision@'
        parameter (string) -- model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int) -- numbers of model group id
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    met_df = (df.loc[(df['metric'] == metric) &
                     (df['parameter'] == parameter)]
              .groupby(['model_group_id'])['raw_value']
              .std())

    if met_df.isnull().sum() == met_df.shape[0]:
        # variance will be undefined in first time window since we only have one obseravtion
        # per model group
        logging.info("Null metric variances for %s %s at %s; picking at random",
                     metric, parameter, train_end_time)
        return df['model_group_id'].drop_duplicates().sample(n=n).tolist()
    elif met_df.isnull().sum() > 0:
        # the variances should be all null or no nulls, a mix shouldn't be possible
        # since we should have the same number of observations for every model group
        raise ValueError("Mix of null and non-null metric variances for or {} {} at {}"
                         .format(metric, parameter, train_end_time))
    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        return [met_df.sample(frac=1).idxmin()]
    else:
        return met_df.nsmallest(n).index.tolist()


def most_frequent_best_dist(df, train_end_time, metric, parameter, dist_from_best_case, n=1):
    """Pick the model that is most frequently within `dist_from_best_case` from the
    best-performing model group across test sets so far

    Arguments:
        dist_from_best_case (float) -- distance from the best performing model
        metric (string) -- model evaluation metric, such as 'precision@'
        parameter (string) -- model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int) -- numbers of model group id
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    met_df = df.loc[
                (df['metric'] == metric) &
                (df['parameter'] == parameter)
            ]
    met_df['within_dist'] = (df['dist_from_best_case'] <= dist_from_best_case).astype('int')
    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        return [met_df.groupby(['model_group_id'])['within_dist'].mean().sample(frac=1).idxmax()]
    else:
        return met_df.groupby(['model_group_id'])['within_dist'].mean().nlargest(n).index.tolist()


def best_average_two_metrics(
    df,
    train_end_time,
    metric1,
    parameter1,
    metric2,
    parameter2,
    metric1_weight=0.5,
    n=1
):
    """Pick the model with the highest average combined value to date
    of two metrics weighted together using `metric1_weight`

    Arguments:
        metric1_weight (float) -- relative weight of metric1, between 0 and 1
        metric1 (string) -- model evaluation metric, such as 'precision@'
        parameter1 (string) -- model evaluation metric parameter,
            such as '300_abs'
        metric2 (string) -- model evaluation metric, such as 'precision@'
        parameter2 (string) -- model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int) -- numbers of model group id
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    if metric1_weight < 0 or metric1_weight > 1:
        raise ValueError("Metric weight must be between 0 and 1")

    metric1_dir = greater_is_better(metric1)
    metric2_dir = greater_is_better(metric2)
    if metric1_dir != metric2_dir:
        raise ValueError("Metric directionalities must be the same")

    met_df = df.loc[
                (
                    (df['metric'] == metric1) &
                    (df['parameter'] == parameter1)
                ) |
                (
                    (df['metric'] == metric2) &
                    (df['parameter'] == parameter2)
                )
            ]

    met_df.loc[
        (met_df['metric'] == metric1) & (met_df['parameter'] == parameter1),
        'weighted_raw'
    ] = met_df.loc[
        (met_df['metric'] == metric1) & (met_df['parameter'] == parameter1),
        'raw_value'
    ] * metric1_weight

    met_df.loc[
        (met_df['metric'] == metric2) & (met_df['parameter'] == parameter2),
        'weighted_raw'
    ] = met_df.loc[
        (met_df['metric'] == metric2) & (met_df['parameter'] == parameter2),
        'raw_value'
    ] * (1.0 - metric1_weight)

    met_df_wt = met_df.groupby(['model_group_id', 'train_end_time'], as_index=False).sum()

    # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
    return _mg_best_avg_by(met_df_wt, 'weighted_raw', metric1, n)


def best_avg_var_penalized(df, train_end_time, metric, parameter, stdev_penalty, n=1):
    """Pick the model with the highest average metric value so far, penalized
    for relative variance as:
        avg_value - (stdev_penalty) * (stdev - min_stdev)
    where min_stdev is the minimum standard deviation of the metric across all
    model groups

    Arguments:
        stdev_penalty (float) -- penalty for instability
        metric (string) -- model evaluation metric, such as 'precision@'
        parameter (string) -- model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int) -- numbers of model group id
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    # for metrics where smaller values are better, the penalty for instability should
    # add to the mean, so introduce a factor of -1
    stdev_penalty = stdev_penalty if greater_is_better(metric) else -1.0*stdev_penalty

    met_df = df.loc[
                (df['metric'] == metric) &
                (df['parameter'] == parameter)
            ]
    met_df_grp = (met_df
                  .groupby(['model_group_id'])
                  .aggregate({'raw_value': ['mean', 'std']}))
    met_df_grp.columns = met_df_grp.columns.droplevel(0)
    met_df_grp.columns = ['raw_avg', 'raw_stdev']
    if met_df_grp['raw_stdev'].isnull().sum() == met_df_grp.shape[0]:
        # variance will be undefined in first time window since we only have one obseravtion
        # per model group
        logging.info("Null metric variances for %s %s at %s; just using mean",
                     metric, parameter, train_end_time)
        return [getattr(
          met_df_grp['raw_avg'].sample(frac=1),
          idxbest(metric)
        )()]
    elif met_df_grp['raw_stdev'].isnull().sum() > 0:
        # the variances should be all null or no nulls, a mix shouldn't be possible
        # since we should have the same number of observations for every model group
        raise ValueError("Mix of null and non-null metric variances for or {} {} at {}"
                         .format(metric, parameter, train_end_time))

    min_stdev = met_df_grp['raw_stdev'].min()
    met_df_grp['penalized_avg'] = (met_df_grp['raw_avg'] -
                                   stdev_penalty * (met_df_grp['raw_stdev'] - min_stdev))

    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        return [getattr(
            met_df_grp['penalized_avg'].sample(frac=1),
            idxbest(metric)
        )()]
    else:
        if greater_is_better(metric):
            return met_df_grp['penalized_avg'].nlargest(n).index.tolist()
        else:
            return met_df_grp['penalized_avg'].nsmallest(n).index.tolist()


def best_avg_recency_weight(df, train_end_time, metric, parameter, curr_weight, decay_type, n=1):
    """Pick the model with the highest average metric value so far, penalized
    for relative variance as:
        avg_value - (stdev_penalty) * (stdev - min_stdev)
    where min_stdev is the minimum standard deviation of the metric across all
    model groups

    Arguments:
        decay_type (string) -- either 'linear' or 'exponential'; the shape of
            how the weights fall off between the current and first point
        curr_weight (float) -- amount of weight to put on the most recent point,
            relative to the first point (e.g., a value of 5.0 would mean the
            current data is weighted 5 times as much as the first one)
        metric (string) -- model evaluation metric, such as 'precision@'
        parameter (string) -- model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp) -- current train end time
        df (pandas.DataFrame) -- dataframe containing the columns:
                model_group_id,
                model_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int) -- numbers of model group id
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    # curr_weight is amount of weight to put on current point, relative to the first point
    # (e.g., if the first point has a weight of 1.0)
    # decay type is linear or exponetial

    first_date = df['train_end_time'].min()
    df['days_out'] = (df['train_end_time'] - first_date).apply(lambda x: float(x.days))
    tmax = df['days_out'].max()

    if tmax == 0:
        # only one date (must be on first time point), so everything gets a weight of 1
        df['weight'] = 1.0
    elif decay_type == 'linear':
        # weight = (curr_weight - 1.0) * (t/tmax) + 1.0
        df['weight'] = (curr_weight - 1.0) * (df['days_out']/tmax) + 1.0
    elif decay_type == 'exponential':
        # weight = exp(ln(curr_weight)*t/tmax)
        df['weight'] = exp(log(curr_weight)*df['days_out']/tmax)
    else:
        raise ValueError('Must specify linear or exponential decay type')

    def wm(x):
        return average(x, weights=df.loc[x.index, 'weight'])

    met_df = df.loc[(df['metric'] == metric) &
                    (df['parameter'] == parameter)]
    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        result = getattr(
            met_df.groupby(['model_group_id']).aggregate({'raw_value': wm}).sample(frac=1),
            idxbest(metric)
        )()
        return result.tolist()
    else:
        met_df_grp = met_df.groupby(['model_group_id']).aggregate({'raw_value': wm})
        if greater_is_better(metric):

            return met_df_grp['raw_value'].nlargest(n).index.tolist()
        else:
            return met_df_grp['raw_value'].nsmallest(n).index.tolist()


SELECTION_RULES = {
    'random_model_group': random_model_group,
    'best_current_value': best_current_value,
    'best_average_value': best_average_value,
    'lowest_metric_variance': lowest_metric_variance,
    'most_frequent_best_dist': most_frequent_best_dist,
    'best_average_two_metrics': best_average_two_metrics,
    'best_avg_var_penalized': best_avg_var_penalized,
    'best_avg_recency_weight': best_avg_recency_weight
}


class BoundSelectionRule(object):
    """A selection rule bound with a set of arguments

    Args:
        args (dict) A set of keyword arguments, that should be sufficient
            to call the function when a dataframe and train_end_time is added
        function_name (string, optional) The name of a function in SELECTION_RULES
        descriptive_name (string, optional) A descriptive name, used in charts
            If none is given it will be automatically constructed
        function (function, optional) A function
    """
    def __init__(self, args, function_name=None, descriptive_name=None, function=None):
        if not function_name and not function:
            raise ValueError('Need either function_name or function')

        if not descriptive_name and not function_name:
            raise ValueError('Need either descriptive_name or function_name')

        self.args = args
        self.function_name = function_name
        self._function = function
        self._descriptive_name = descriptive_name

    @property
    def function(self):
        if not self._function:
            self._function = SELECTION_RULES[self.function_name]
        return self._function

    @property
    def descriptive_name(self):
        if not self._descriptive_name:
            self._descriptive_name = self._build_descriptive_name()

        return self._descriptive_name

    def __str__(self):
        return self.descriptive_name

    def _build_descriptive_name(self):
        """Build a descriptive name for the bound selection rule

        Constructed using the function name and arguments.
        """
        argspec = inspect.getargspec(self.function)
        args = [
            arg for arg in argspec.args
            if arg not in ['df', 'train_end_time', 'n']
        ]
        return '_'.join(
            [self.function_name] +
            [str(self.args[key]) for key in args]
        )

    def pick(self, dataframe, train_end_time):
        """Run the selection rule for a given time on a dataframe

        Args:
            dataframe (pandas.DataFrame)
            train_end_time (timestamp) Current train end time

        Returns: (int) a model group id
        """
        return self.function(dataframe, train_end_time, **(self.args))



import warnings

import numpy as np
import sklearn.linear_model
from sklearn.svm import SVC


def _ad_hoc_feature_importances(model):
    """
    Get the "ad-hoc feature importances" for scikit-learn's models
    lacking the `feature_importances_` attribute

    Args:
        model: A trained model that has not a `feature_importances_` attribute

    Returns:
        At this moment, this method only returns the odds ratio of both the
        intercept and the coefficients given by sklearn's implementation of
        the LogisticRegression.
        The order of the odds ratio list is the standard
        of the statistical packages (like R, SAS, etc) i.e. (intercept, coefficients)
    """
    feature_importances = None

    if isinstance(model, (sklearn.linear_model.logistic.LogisticRegression)):
        coef_odds_ratio = np.exp(model.coef_)
        # intercept_odds_ratio = np.exp(model.intercept_[:,np.newaxis])
        # We are ignoring the intercept

        # NOTE: We need to squeeze this array so it has the correct dimensions
        feature_importances = coef_odds_ratio.squeeze()

    return feature_importances


def get_feature_importances(model):
    """
    Get feature importances (from scikit-learn) of a trained model.

    Args:
        model: Trained model

    Returns:
        Feature importances, or failing that, None
    """
    feature_importances = None

    if hasattr(model, 'feature_importances_'):
        feature_importances = model.feature_importances_

    elif isinstance(model, (SVC)) and (model.get_params()['kernel'] == 'linear'):
        feature_importances = model.coef_.squeeze()

    else:
        warnings.warn(
            "\nThe selected algorithm, doesn't support a standard way"
            "\nof calculate the importance of each feature used."
            "\nFalling back to ad-hoc methods"
            "\n(e.g. in LogisticRegression we will return Odd Ratios instead coefficients)"
        )

        feature_importances = _ad_hoc_feature_importances(model)

    return feature_importances



# coding: utf-8

import sqlalchemy
import wrapt


class SerializableDbEngine(wrapt.ObjectProxy):
    """A sqlalchemy engine that can be serialized across process boundaries.

    Works by saving all kwargs used to create the engine and reconstructs them later.
    As a result, the state won't be saved upon serialization/deserialization.
    """

    __slots__ = ('url', 'creator', 'kwargs')

    def __init__(self, url, *, creator=sqlalchemy.create_engine, **kwargs):
        self.url = url
        self.creator = creator
        self.kwargs = kwargs

        engine = creator(url, **kwargs)
        super().__init__(engine)

    def __reduce__(self):
        return (self.__reconstruct__, (self.url, self.creator, self.kwargs))

    @classmethod
    def __reconstruct__(cls, url, creator, kwargs):
        return cls(url, creator=creator, **kwargs)


create_engine = SerializableDbEngine



from datetime import datetime
import tempfile
import yaml
import os
import mock

import factory
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition import Auditioner, AuditionRunner
from triage.component.catwalk.db import ensure_db

from tests.results_tests.factories import (
    EvaluationFactory,
    ModelFactory,
    ModelGroupFactory,
    init_engine,
    session,
    MatrixFactory
)

config = {
    'filter': {'distance_table': 'distance_table',
    'max_from_best': 1.0,
    'metric': 'precision@',
    'models_table': 'models',
    'parameter': '50_abs',
    'threshold_value': 0.0},
    'model_groups': {'query': 'SELECT DISTINCT(model_group_id) FROM model_metadata.model_groups'},
    'rules':
        [
            {
                'selection_rules': [
                    {'name': 'best_current_value', 'n': 3},
                    {'name': 'best_average_value', 'n': 3},
                    {'name': 'lowest_metric_variance', 'n': 3},
                    {'dist_from_best_case': [0.05], 'name': 'most_frequent_best_dist', 'n': 3}
                    ],
                'shared_parameters': [
                    {'metric': 'precision@', 'parameter': '50_abs'},
                    ]
            },
            {
                'selection_rules': [
                    {'metric1_weight': 0.5, 'metric2': 'recall@','name': 'best_average_two_metrics', 'n': 3, 'parameter2': '50_abs'}
                    ],
                'shared_parameters': [
                    {'metric1': 'precision@', 'parameter1': '50_abs'}
                    ]
            }
        ],
    'time_stamps': {'query': "SELECT DISTINCT train_end_time FROM model_metadata.models WHERE model_group_id IN ({}) AND EXTRACT(DAY FROM train_end_time) IN (1) AND train_end_time >= '2012-01-01'"}
}

@mock.patch('os.getcwd')
def test_Audition(mock_getcwd):
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        init_engine(db_engine)

        num_model_groups = 10
        model_types = [
            'classifier type {}'.format(i)
            for i in range(0, num_model_groups)
        ]
        model_groups = [
            ModelGroupFactory(model_type=model_type)
            for model_type in model_types
        ]
        train_end_times = [
            datetime(2013, 1, 1),
            datetime(2014, 1, 1),
            datetime(2015, 1, 1),
            datetime(2016, 1, 1),
        ]

        models = [
            ModelFactory(model_group_rel=model_group, train_end_time=train_end_time)
            for model_group in model_groups
            for train_end_time in train_end_times
        ]
        metrics = [
            ('precision@', '100_abs'),
            ('recall@', '100_abs'),
            ('precision@', '50_abs'),
            ('recall@', '50_abs'),
            ('fpr@', '10_pct'),
        ]

        class ImmediateEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(lambda o: o.model_rel.train_end_time)

        for model in models:
            for (metric, parameter) in metrics:
                ImmediateEvalFactory(model_rel=model,
                                     metric=metric,
                                     parameter=parameter)

        session.commit()

        with tempfile.TemporaryDirectory() as td:
            mock_getcwd.return_value = td
            AuditionRunner(config_dict=config, db_engine=db_engine, directory=td).run()
            assert len(os.listdir(os.getcwd())) == 6


def test_Auditioner():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        init_engine(db_engine)
        # set up data, randomly generated by the factories but conforming
        # generally to what we expect model_metadata schema data to look like

        num_model_groups = 10
        model_types = [
            'classifier type {}'.format(i)
            for i in range(0, num_model_groups)
        ]
        model_groups = [
            ModelGroupFactory(model_type=model_type)
            for model_type in model_types
        ]
        train_end_times = [
            datetime(2013, 1, 1),
            datetime(2014, 1, 1),
            datetime(2015, 1, 1),
            datetime(2016, 1, 1),
        ]

        models = [
            ModelFactory(model_group_rel=model_group, train_end_time=train_end_time)
            for model_group in model_groups
            for train_end_time in train_end_times
        ]
        metrics = [
            ('precision@', '100_abs'),
            ('recall@', '100_abs'),
            ('precision@', '50_abs'),
            ('recall@', '50_abs'),
            ('fpr@', '10_pct'),
        ]

        class ImmediateEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(lambda o: o.model_rel.train_end_time)

        for model in models:
            for (metric, parameter) in metrics:
                ImmediateEvalFactory(model_rel=model,
                                     metric=metric,
                                     parameter=parameter)

        session.commit()

        # define a very loose filtering that should admit all model groups
        no_filtering = [
            {
                'metric': 'precision@',
                'parameter': '100_abs',
                'max_from_best': 1.0,
                'threshold_value': 0.0
            },
            {
                'metric': 'recall@',
                'parameter': '100_abs',
                'max_from_best': 1.0,
                'threshold_value': 0.0
            }
        ]
        model_group_ids = [mg.model_group_id for mg in model_groups]
        auditioner = Auditioner(
            db_engine,
            model_group_ids,
            train_end_times,
            no_filtering,
        )
        assert len(auditioner.thresholded_model_group_ids) == num_model_groups
        auditioner.plot_model_groups()


        # here, we pick thresholding rules that should definitely remove
        # all model groups from contention because they are too strict.
        remove_all = [
            {
                'metric': 'precision@',
                'parameter': '100_abs',
                'max_from_best': 0.0,
                'threshold_value': 1.1
            },
            {
                'metric': 'recall@',
                'parameter': '100_abs',
                'max_from_best': 0.0,
                'threshold_value': 1.1
            }
        ]

        auditioner.update_metric_filters(new_filters=remove_all)
        assert len(auditioner.thresholded_model_group_ids) == 0

        # pass the argument instead and remove all model groups
        auditioner.set_one_metric_filter(
            metric='precision@',
            parameter='100_abs',
            max_from_best=0.0,
            threshold_value=1.1)
        assert len(auditioner.thresholded_model_group_ids) == 0

        # one potential place for bugs would be when we pull back the rules
        # for being too restrictive. we want to make sure that the original list is
        # always used for thresholding, or else such a move would be impossible
        auditioner.update_metric_filters(new_filters=no_filtering)
        assert len(auditioner.thresholded_model_group_ids) == num_model_groups

        # pass the argument instead and let all model groups pass
        auditioner.set_one_metric_filter(
            metric='precision@',
            parameter='100_abs',
            max_from_best=1.0,
            threshold_value=0.0)
        assert len(auditioner.thresholded_model_group_ids) == num_model_groups

        # now, we want to take this partially thresholded list and run it through
        # a grid of selection rules, meant to pick winners by a variety of user-defined
        # criteria
        rule_grid = [{
            'shared_parameters': [
                {'metric': 'precision@', 'parameter': '100_abs'},
                {'metric': 'recall@', 'parameter': '100_abs'},
            ],
            'selection_rules': [
                {'name': 'most_frequent_best_dist', 'dist_from_best_case': [0.1, 0.2, 0.3], 'n': 1},
                {'name': 'best_current_value', 'n': 1}
            ]
        }, {
            'shared_parameters': [
                {'metric1': 'precision@', 'parameter1': '100_abs'},
            ],
            'selection_rules': [
                {
                    'name': 'best_average_two_metrics',
                    'metric2': ['recall@'],
                    'parameter2': ['100_abs'],
                    'metric1_weight': [0.4, 0.5, 0.6],
                    'n': 1
                },
            ]
        }]
        auditioner.register_selection_rule_grid(rule_grid, plot=False)
        final_model_group_ids = auditioner.selection_rule_model_group_ids

        # we expect the result to be a mapping of selection rule name to model group id
        assert isinstance(final_model_group_ids, dict)

        # we expect that there is one winner for each selection rule
        assert sorted(final_model_group_ids.keys()) == \
            sorted([rule.descriptive_name for rule in auditioner.selection_rules])

        # we expect that the results written to the yaml file are the
        # chosen model groups and their rules
        # however because the source data is randomly generated we could have a
        # different list on consecutive runs
        # and don't want to introduce non-determinism to the test
        with tempfile.NamedTemporaryFile() as tf:
            auditioner.write_tyra_config(tf.name)
            assert sorted(yaml.load(tf)['selection_rule_model_groups'].keys()) == \
                sorted(final_model_group_ids.keys())



"""empty message

Revision ID: 264245ddfce2
Revises: 0d44655e35fd
Create Date: 2017-09-01 14:26:01.107455

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '264245ddfce2'
down_revision = '0d44655e35fd'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'individual_importances',
        sa.Column('feature_value', sa.Float(), nullable=True),
        schema='results',
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('individual_importances', 'feature_value', schema='results')
    # ### end Alembic commands ###



from contextlib import contextmanager
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import make_transient
import datetime
from unittest.mock import Mock
from numpy.testing import assert_array_equal
import pandas

from triage.component.results_schema import TestPrediction, Matrix, Model

from triage.component.catwalk.predictors import Predictor
from tests.utils import MockTrainedModel, matrix_creator, matrix_metadata_creator, get_matrix_store, rig_engines


AS_OF_DATE = datetime.date(2016, 12, 21)


@contextmanager
def prepare():
    with rig_engines() as (db_engine, project_storage):
        train_matrix_uuid = '1234'
        session = sessionmaker(db_engine)()
        session.add(Matrix(matrix_uuid=train_matrix_uuid))

        # Create the fake trained model and store in db
        trained_model = MockTrainedModel()
        model_hash = 'abcd'
        project_storage.model_storage_engine().write(trained_model, model_hash)
        db_model = Model(model_hash=model_hash, train_matrix_uuid=train_matrix_uuid)
        session.add(db_model)
        session.commit()
        yield project_storage, db_engine, db_model.model_id


def test_predictor_entity_index():
    with prepare() as (project_storage, db_engine, model_id):
        predictor = Predictor(project_storage.model_storage_engine(), db_engine)

        # Runs the same test for training and testing predictions
        for mat_type in ("train", "test"):
            matrix = matrix_creator(index='entity_id')
            metadata = matrix_metadata_creator(end_time=AS_OF_DATE, matrix_type=mat_type, indices=['entity_id'])

            matrix_store = get_matrix_store(project_storage, matrix, metadata)
            train_matrix_columns = matrix.columns[0:-1].tolist()

            predict_proba = predictor.predict(
                model_id,
                matrix_store,
                misc_db_parameters=dict(),
                train_matrix_columns=train_matrix_columns
            )

            # assert
            # 1. that the returned predictions are of the desired length
            assert len(predict_proba) == 2

            # 2. that the predictions table entries are present and
            # can be linked to the original models
            records = [
                row for row in
                db_engine.execute('''select entity_id, as_of_date
                from {}_results.predictions
                join model_metadata.models using (model_id)'''.format(mat_type, mat_type))
            ]
            assert len(records) == 2

            # 3. that the contained as_of_dates match what we sent in
            for record in records:
                assert record[1].date() == AS_OF_DATE

            # 4. that the entity ids match the given dataset
            assert sorted([record[0] for record in records]) == [1, 2]

        # 5. running with same model_id, different as of date
        # then with same as of date only replaces the records
        # with the same date

        # Runs the same test for training and testing predictions
        for mat_type in ("train", "test"):
            new_matrix = matrix_creator(index='entity_id')
            new_metadata = matrix_metadata_creator(end_time=AS_OF_DATE + datetime.timedelta(days=1), matrix_type=mat_type, indices=['entity_id'])
            new_matrix_store = get_matrix_store(project_storage, new_matrix, new_metadata)

            predictor.predict(
                model_id,
                new_matrix_store,
                misc_db_parameters=dict(),
                train_matrix_columns=train_matrix_columns
            )
            predictor.predict(
                model_id,
                matrix_store,
                misc_db_parameters=dict(),
                train_matrix_columns=train_matrix_columns
            )
            records = [
                row for row in
                db_engine.execute('''select entity_id, as_of_date
                from {}_results.predictions
                join model_metadata.models using (model_id)'''.format(mat_type, mat_type))
            ]
            assert len(records) == 4

        # 6. That we can delete the model when done prediction on it
        predictor.delete_model(model_id)
        assert predictor.load_model(model_id) is None


def test_predictor_composite_index():
    with prepare() as (project_storage, db_engine, model_id):
        predictor = Predictor(project_storage.model_storage_engine(), db_engine)

        dayone = datetime.datetime(2011, 1, 1)
        daytwo = datetime.datetime(2011, 1, 2)
        source_dict = {
            'entity_id': [1, 2, 1, 2],
            'as_of_date': [dayone, dayone, daytwo, daytwo],
            'feature_one': [3, 4, 5, 6],
            'feature_two': [5, 6, 7, 8],
            'label': [7, 8, 8, 7]
        }

        # Runs the same test for training and testing predictions
        for mat_type in ("train", "test"):

            matrix = pandas.DataFrame.from_dict(source_dict).set_index(['entity_id', 'as_of_date'])
            metadata = matrix_metadata_creator(matrix_type=mat_type)
            matrix_store = get_matrix_store(project_storage, matrix, metadata)

            predict_proba = predictor.predict(
                model_id,
                matrix_store,
                misc_db_parameters=dict(),
                train_matrix_columns=['feature_one', 'feature_two']
            )

            # assert
            # 1. that the returned predictions are of the desired length
            assert len(predict_proba) == 4

            # 2. that the predictions table entries are present and
            # can be linked to the original models
            records = [
                row for row in
                db_engine.execute('''select entity_id, as_of_date
                from {}_results.predictions
                join model_metadata.models using (model_id)'''.format(mat_type, mat_type))
            ]
            assert len(records) == 4


def test_predictor_get_train_columns():
    with prepare() as (project_storage, db_engine, model_id):
        predictor = Predictor(project_storage.model_storage_engine(), db_engine)
        train_store = get_matrix_store(
            project_storage=project_storage,
            matrix=matrix_creator(),
            metadata=matrix_metadata_creator(matrix_type='train'),
        )

        # flip the order of some feature columns in the test matrix
        other_order_matrix = matrix_creator()
        order = other_order_matrix.columns.tolist()
        order[0], order[1] = order[1], order[0]
        other_order_matrix = other_order_matrix[order]
        test_store = get_matrix_store(
            project_storage=project_storage,
            matrix=other_order_matrix,
            metadata=matrix_metadata_creator(matrix_type='test'),
        )

        # Runs the same test for training and testing predictions
        for store, mat_type in zip((train_store, test_store), ("train", "test")):
            predict_proba = predictor.predict(
                model_id,
                store,
                misc_db_parameters=dict(),
                train_matrix_columns=train_store.columns()
            )
            # assert
            # 1. that we calculated predictions
            assert len(predict_proba) > 0

            # 2. that the predictions table entries are present and
            # can be linked to the original models
            records = [
                row for row in
                db_engine.execute('''select entity_id, as_of_date
                from {}_results.predictions
                join model_metadata.models using (model_id)'''.format(mat_type, mat_type))
            ]
            assert len(records) > 0


def test_predictor_retrieve():
    with prepare() as (project_storage, db_engine, model_id):
        predictor = Predictor(project_storage.model_storage_engine(), db_engine, replace=False)

        # create prediction set
        matrix = matrix_creator()
        metadata = matrix_metadata_creator()
        matrix_store = get_matrix_store(project_storage, matrix, metadata)

        predict_proba = predictor.predict(
            model_id,
            matrix_store,
            misc_db_parameters=dict(),
            train_matrix_columns=matrix.columns[0:-1].tolist()
        )

        # When run again, the predictions retrieved from the database
        # should match.
        #
        # Some trickiness here. Let's explain:
        #
        # If we are not careful, retrieving predictions from the database and
        # presenting them as a numpy array can result in a bad ordering,
        # since the given matrix may not be 'ordered' by some criteria
        # that can be easily represented by an ORDER BY clause.
        #
        # It will sometimes work, because without ORDER BY you will get
        # it back in the table's physical order, which unless something has
        # happened to the table will be the order you inserted it,
        # which could very well be the order in the matrix.
        # So it's not a bug that would necessarily immediately show itself,
        # but when it does go wrong your scores will be garbage.
        #
        # So we simulate a table order mutation that can happen over time:
        # Remove the first row and put it at the end.
        # If the Predictor doesn't explicitly reorder the results, this will fail
        # Only running on TestPrediction because TrainPrediction behaves the exact same way
        reorder_session = sessionmaker(bind=db_engine)()
        obj = reorder_session.query(TestPrediction).first()
        reorder_session.delete(obj)
        reorder_session.commit()

        make_transient(obj)
        reorder_session = sessionmaker(bind=db_engine)()
        reorder_session.add(obj)
        reorder_session.commit()

        predictor.load_model = Mock()
        new_predict_proba = predictor.predict(
            model_id,
            matrix_store,
            misc_db_parameters=dict(),
            train_matrix_columns=matrix.columns[0:-1].tolist()
        )
        assert_array_equal(new_predict_proba, predict_proba)
        assert not predictor.load_model.called



from triage.component.catwalk.evaluation import ModelEvaluator, generate_binary_at_x
from triage.component.catwalk.metrics import Metric
import testing.postgresql

import numpy
from sqlalchemy import create_engine
from triage.component.catwalk.db import ensure_db
from tests.utils import fake_labels, fake_trained_model, MockMatrixStore
import datetime


@Metric(greater_is_better=True)
def always_half(predictions_proba, predictions_binary, labels, parameters):
    return 0.5


def test_evaluating_early_warning():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        testing_metric_groups = [{
            'metrics': ['precision@',
                        'recall@',
                        'true positives@',
                        'true negatives@',
                        'false positives@',
                        'false negatives@'],
            'thresholds': {
                'percentiles': [5.0, 10.0],
                'top_n': [5, 10]
            }
        }, {
            'metrics': ['f1',
                        'mediocre',
                        'accuracy',
                        'roc_auc',
                        'average precision score'],
        }, {
            'metrics': ['fbeta@'],
            'parameters': [{'beta': 0.75}, {'beta': 1.25}]
        }]

        training_metric_groups = [{'metrics': ['accuracy', 'roc_auc']}]

        custom_metrics = {'mediocre': always_half}

        model_evaluator = ModelEvaluator(testing_metric_groups, training_metric_groups, db_engine,
            custom_metrics=custom_metrics
        )

        labels = fake_labels(5)
        fake_train_matrix_store = MockMatrixStore('train', 'efgh', 5, db_engine, labels)
        fake_test_matrix_store = MockMatrixStore('test', '1234', 5, db_engine, labels)

        trained_model, model_id = fake_trained_model(db_engine)

        # Evaluate the testing metrics and test for all of them.
        model_evaluator.evaluate(
            trained_model.predict_proba(labels)[:, 1],
            fake_test_matrix_store,
            model_id,
        )
        records = [
            row[0] for row in
            db_engine.execute(
                '''select distinct(metric || parameter)
                from test_results.evaluations
                where model_id = %s and
                evaluation_start_time = %s
                order by 1''',
                (model_id, fake_test_matrix_store.as_of_dates[0])
            )
        ]
        assert records == [
            'accuracy',
            'average precision score',
            'f1',
            'false negatives@10.0_pct',
            'false negatives@10_abs',
            'false negatives@5.0_pct',
            'false negatives@5_abs',
            'false positives@10.0_pct',
            'false positives@10_abs',
            'false positives@5.0_pct',
            'false positives@5_abs',
            'fbeta@0.75_beta',
            'fbeta@1.25_beta',
            'mediocre',
            'precision@10.0_pct',
            'precision@10_abs',
            'precision@5.0_pct',
            'precision@5_abs',
            'recall@10.0_pct',
            'recall@10_abs',
            'recall@5.0_pct',
            'recall@5_abs',
            'roc_auc',
            'true negatives@10.0_pct',
            'true negatives@10_abs',
            'true negatives@5.0_pct',
            'true negatives@5_abs',
            'true positives@10.0_pct',
            'true positives@10_abs',
            'true positives@5.0_pct',
            'true positives@5_abs'
        ]

        # Evaluate the training metrics and test
        model_evaluator.evaluate(
            trained_model.predict_proba(labels)[:, 1],
            fake_train_matrix_store,
            model_id,
        )
        records = [
            row[0] for row in
            db_engine.execute(
                '''select distinct(metric || parameter)
                from train_results.evaluations
                where model_id = %s and
                evaluation_start_time = %s
                order by 1''',
                (model_id, fake_train_matrix_store.as_of_dates[0])
            )
        ]
        assert records == ['accuracy', 'roc_auc']


def test_model_scoring_inspections():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        testing_metric_groups = [{
            'metrics': ['precision@', 'recall@', 'fpr@'],
            'thresholds': {'percentiles': [50.0], 'top_n': [3]}
        }, {
            # ensure we test a non-thresholded metric as well
            'metrics': ['accuracy'],
        }]
        training_metric_groups = [{'metrics': ['accuracy'], 'thresholds': {'percentiles': [50.0]}}]

        model_evaluator = ModelEvaluator(testing_metric_groups, training_metric_groups, db_engine)

        testing_labels = numpy.array([True, False, numpy.nan, True, False])
        testing_prediction_probas = numpy.array([0.56, 0.4, 0.55, 0.5, 0.3])

        training_labels = numpy.array([False, False, True, True, True, False, True, True])
        training_prediction_probas = numpy.array([0.6, 0.4, 0.55, 0.70, 0.3, 0.2, 0.8, 0.6])

        fake_train_matrix_store = MockMatrixStore('train', 'efgh', 5, db_engine, training_labels)
        fake_test_matrix_store = MockMatrixStore('test', '1234', 5, db_engine, testing_labels)

        trained_model, model_id = fake_trained_model(db_engine)

        # Evaluate testing matrix and test the results
        model_evaluator.evaluate(
            testing_prediction_probas,
            fake_test_matrix_store,
            model_id,
        )
        for record in db_engine.execute(
            '''select * from test_results.evaluations
            where model_id = %s and evaluation_start_time = %s
            order by 1''',
            (model_id, fake_test_matrix_store.as_of_dates[0])
        ):
            assert record['num_labeled_examples'] == 4
            assert record['num_positive_labels'] == 2
            if record['parameter'] == '':
                assert record['num_labeled_above_threshold'] == 4
            elif 'pct' in record['parameter']:
                assert record['num_labeled_above_threshold'] == 1
            else:
                assert record['num_labeled_above_threshold'] == 2

        # Evaluate the training matrix and test the results
        model_evaluator.evaluate(
                    training_prediction_probas,
                    fake_train_matrix_store,
                    model_id,
        )
        for record in db_engine.execute(
            '''select * from train_results.evaluations
            where model_id = %s and evaluation_start_time = %s
            order by 1''',
            (model_id, fake_train_matrix_store.as_of_dates[0])
        ):
            assert record['num_labeled_examples'] == 8
            assert record['num_positive_labels'] == 5
            assert record['value'] == 0.625


def test_generate_binary_at_x():
    input_list = [0.9, 0.8, 0.7, 0.7, 0.7, 0.7, 0.7, 0.7, 0.7, 0.6]

    # bug can arise when the same value spans both sides of threshold
    assert generate_binary_at_x(input_list, 50, 'percentile') == \
        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0]

    assert generate_binary_at_x(input_list, 2) == \
        [1, 1, 0, 0, 0, 0, 0, 0, 0, 0]



from .metta_io import (
    archive_matrix,
    generate_uuid,
    archive_train_test,
)
from .upload_s3 import upload_to_s3

__all__ = (
    'archive_matrix',
    'archive_train_test',
    'generate_uuid',
    'upload_to_s3',
)



import numpy

import warnings

import pytest

from triage.component.catwalk.feature_importances import _ad_hoc_feature_importances, get_feature_importances

from sklearn import datasets
from sklearn.svm import SVC
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression, LinearRegression

from sklearn.model_selection import train_test_split

@pytest.fixture
def trained_models():
    dataset = datasets.load_breast_cancer()
    X = dataset.data
    y = dataset.target

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = 0.3, random_state=12345)

    rf = RandomForestClassifier()
    rf.fit(X_train, y_train)

    lr = LogisticRegression()
    lr.fit(X_train, y_train)

    svc_w_linear_kernel = SVC(kernel='linear')
    svc_w_linear_kernel.fit(X_train, y_train)

    svc_wo_linear_kernel = SVC()
    svc_wo_linear_kernel.fit(X_train, y_train)

    dummy = DummyClassifier()
    dummy.fit(X_train, y_train)

    return {'RF':rf, 'LR':lr, 'SVC_w_linear_kernel':svc_w_linear_kernel,
            'Dummy':dummy, 'SVC_wo_linear_kernel':svc_wo_linear_kernel}

def test_throwing_warning_if_lr(trained_models):
    with pytest.warns(UserWarning):
        get_feature_importances(trained_models['LR'])

def test_throwing_warning_if_dummyclassifier(trained_models):
    with pytest.warns(UserWarning):
        get_feature_importances(trained_models['Dummy'])

def test_throwing_warning_if_SVC_wo_linear_kernel(trained_models):
    with pytest.warns(UserWarning):
        get_feature_importances(trained_models['SVC_wo_linear_kernel'])

def test_correct_feature_importances_for_lr(trained_models):
    feature_importances = get_feature_importances(trained_models['LR'])

    ## It returns the intercept, too
    assert feature_importances.shape == (30,)

def test_correct_feature_importances_for_rf(trained_models):
    feature_importances = get_feature_importances(trained_models['RF'])
    assert feature_importances.shape == (30,)

def test_correct_feature_importances_for_svc_w_linear_kernel(trained_models):
    feature_importances = get_feature_importances(trained_models['SVC_w_linear_kernel'])
    assert feature_importances.shape == (30,)

def test_correct_feature_importances_for_svc_wo_linear_kernel(trained_models):
    feature_importances = get_feature_importances(trained_models['SVC_wo_linear_kernel'])
    assert feature_importances is None

def test_correct_feature_importances_for_dummy(trained_models):
    feature_importances = get_feature_importances(trained_models['Dummy'])
    assert feature_importances is None



import datetime
import shutil
import sys
import tempfile
import random
from contextlib import contextmanager

import pandas as pd
import yaml
import numpy
from sqlalchemy.orm import sessionmaker

from triage.component.results_schema import Model


def convert_string_column_to_date(column):
    return(
        [datetime.datetime.strptime(date, '%Y-%m-%d').date() for date in column]
    )


def create_schemas(engine, features_tables, labels, states):
    """ This function makes a features schema and populates it with the fake
    data from above.

    :param engine: a postgresql engine
    :type engine: sqlalchemy.engine
    """
    # create features schema and associated tables
    engine.execute('drop schema if exists features cascade; create schema features;')
    for table_number, table in enumerate(features_tables):
        create_features_table(table_number, table, engine)
    # create labels schema and table
    engine.execute('drop schema if exists labels cascade; create schema labels;')
    engine.execute(
        """
            create table labels.labels (
                entity_id int,
                as_of_date date,
                label_timespan interval,
                label_name char(30),
                label_type char(30),
                label int
            )
        """
    )
    for row in labels:
        engine.execute(
            'insert into labels.labels values (%s, %s, %s, %s, %s, %s)',
            row
        )

    # create sparse state table
    engine.execute('drop schema if exists staging cascade; create schema staging;')
    engine.execute(
        """
            create table staging.sparse_states (
                entity_id int,
                as_of_date date,
                state_one bool,
                state_two bool
            )
        """
    )
    for row in states:
        engine.execute(
            'insert into staging.sparse_states values (%s, %s, %s, %s)',
            row
        )



def create_features_table(table_number, table, engine):
    engine.execute(
        """
            create table features.features{} (
                entity_id int, as_of_date date, f{} int, f{} int
            )
        """.format(table_number, (table_number*2)+1, (table_number*2)+2)
    )
    for row in table:
        engine.execute(
            """
                insert into features.features{} values (%s, %s, %s, %s)
            """.format(table_number),
            row
        )


def create_entity_date_df(
    labels,
    states,
    as_of_dates,
    state_one,
    state_two,
    label_name,
    label_type,
    label_timespan
):
    """ This function makes a pandas DataFrame that mimics the entity-date table
    for testing against.
    """
    0, '2016-02-01', '1 month', 'booking', 'binary', 0
    labels_table = pd.DataFrame(labels, columns=[
        'entity_id',
        'as_of_date',
        'label_timespan',
        'label_name',
        'label_type',
        'label'
    ])
    states_table = pd.DataFrame(states, columns = [
        'entity_id',
        'as_of_date',
        'state_one',
        'state_two'
    ]).set_index(['entity_id', 'as_of_date'])
    as_of_dates = [date.date() for date in as_of_dates]
    labels_table = labels_table[labels_table['label_name'] == label_name]
    labels_table = labels_table[labels_table['label_type'] == label_type]
    labels_table = labels_table[labels_table['label_timespan'] == label_timespan]
    labels_table = labels_table.join(
        other=states_table,
        on=('entity_id', 'as_of_date'),
    )
    labels_table = labels_table[labels_table['state_one'] & labels_table['state_two']]
    ids_dates = labels_table[['entity_id', 'as_of_date']]
    ids_dates = ids_dates.sort_values(['entity_id', 'as_of_date'])
    ids_dates['as_of_date'] = [datetime.datetime.strptime(
        date,
        '%Y-%m-%d'
    ).date() for date in ids_dates['as_of_date']]
    ids_dates = ids_dates[ids_dates['as_of_date'].isin(as_of_dates)]
    print(ids_dates)

    return(ids_dates.reset_index(drop=True))


def NamedTempFile():
    if sys.version_info >= (3, 0, 0):
        return tempfile.NamedTemporaryFile(mode='w+', newline='')
    else:
        return tempfile.NamedTemporaryFile()


@contextmanager
def TemporaryDirectory():
    name = tempfile.mkdtemp()
    try:
        yield name
    finally:
        shutil.rmtree(name)


@contextmanager
def fake_metta(matrix_dict, metadata):
    """Stores matrix and metadata in a metta-data-like form

    Args:
    matrix_dict (dict) of form { columns: values }.
        Expects an entity_id to be present which it will use as the index
    metadata (dict). Any metadata that should be set

    Yields:
        tuple of filenames for matrix and metadata
    """
    matrix = pd.DataFrame.from_dict(matrix_dict).set_index('entity_id')
    with tempfile.NamedTemporaryFile() as matrix_file:
        with tempfile.NamedTemporaryFile('w') as metadata_file:
            hdf = pd.HDFStore(matrix_file.name)
            hdf.put('title', matrix, data_columns=True)
            matrix_file.seek(0)

            yaml.dump(metadata, metadata_file)
            metadata_file.seek(0)
            yield (matrix_file.name, metadata_file.name)


def fake_labels(length):
    return numpy.array([random.choice([True, False]) for i in range(0, length)])


def assert_index(engine, table, column):
    """Assert that a table has an index on a given column

    Does not care which position the column is in the index
    Modified from https://www.gab.lc/articles/index_on_id_with_postgresql

    Args:
        engine (sqlalchemy.engine) a database engine
        table (string) the name of a table
        column (string) the name of a column
    """
    query = """
        SELECT 1
        FROM pg_class t
             JOIN pg_index ix ON t.oid = ix.indrelid
             JOIN pg_class i ON i.oid = ix.indexrelid
             JOIN pg_attribute a ON a.attrelid = t.oid
        WHERE
             a.attnum = ANY(ix.indkey) AND
             t.relkind = 'r' AND
             t.relname = '{table_name}' AND
             a.attname = '{column_name}'
    """.format(
        table_name=table,
        column_name=column
    )
    num_results = len([row for row in engine.execute(query)])
    assert num_results >= 1


def create_dense_state_table(db_engine, table_name, data):
    db_engine.execute('''create table {} (
        entity_id int,
        state text,
        start_time timestamp,
        end_time timestamp
    )'''.format(table_name))

    for row in data:
        db_engine.execute(
            'insert into {} values (%s, %s, %s, %s)'.format(table_name),
            row
        )


def create_binary_outcome_events(db_engine, table_name, events_data):
    db_engine.execute(
        'create table events (entity_id int, outcome_date date, outcome bool)'
    )
    for event in events_data:
        db_engine.execute(
            'insert into {} values (%s, %s, %s::bool)'.format(table_name),
            event
        )
